class Meta():
    def meta():
        metaString = '''
            query inventoryItems(
            $topLevelField: String!
            $topLevelType: String!
            $nestingLevel: Int! = 3
            $listNestingLevel: Int = 1
            $nestingPlusOneFields: [String!]
            $variables: [VariableValueInput!]
            ) {
            meta {
                query {
                dynamicItem(
                    topLevelField: $topLevelField
                    nestingLevel: $nestingLevel
                    listNestingLevel: $listNestingLevel
                    nestingPlusOneFields: $nestingPlusOneFields
                ) {
                    result(variables: $variables) {
                    grid {
                        columns {
                        graphQlPath
                        path
                        type {
                            name
                            isNonNullable
                        }
                        kind
                        }
                        rows {
                        values {
                            ... on StringCellValue {
                            stringValue: value
                            }
                            ... on BooleanCellValue {
                            boolValue: value
                            }
                            ... on DecimalCellValue {
                            doubleValue: value
                            }
                            ... on IntCellValue {
                            intValue: value
                            }
                            ... on DateTimeCellValue {
                            dateTimeValue: value
                            }
                        }
                        } 
                    }
                    }
                }
                } 
                typeDescription(typeName: $topLevelType) {
                ...typeDescription
                }
            }

            inventories: inventories(where: {name: {eq:  $topLevelField}}) {
                edges {
                node {
                    displayName
                    properties {
                    name  
                    __typename
                    }
                }
                }
            }
            }



            fragment typeDescription on ITypeDescription {
            ...typeDescriptionFields
            ... on ListTypeDescription {
                innerType {
                ...typeDescriptionFields
                ... on ListTypeDescription {
                    innerType {
                    ...typeDescriptionFields
                    }
                }
                ... on ObjectTypeDescription {
                    fields {
                    name
                    type {
                        ...typeDescriptionFields
                    }
                    }
                }
                ... on EnumTypeDescription {
                    ...enumFields
                }    
                }
            }
            ... on ObjectTypeDescription {
                fields {
                name
                type {
                    ...typeDescriptionFields
                    ... on ListTypeDescription {
                    innerType {
                        ...typeDescriptionFields           
                    }
                    }
                    ... on ObjectTypeDescription {
                    fields {
                        name
                        type {
                        ...typeDescriptionFields
                        }
                    }
                    } 
                    ... on EnumTypeDescription {
                    ...enumFields
                    }
                }
                }
            }
            ... on EnumTypeDescription {
                ...enumFields
            }
            }

            fragment typeDescriptionFields on ITypeDescription {
            typeKind: __typename
            name 
            isNonNullable
            }

            fragment enumFields on EnumTypeDescription {
                enumValues: values
            }

            '''
        return metaString



import copy
import pandas as pd

from loguru import logger
from time import sleep
from datetime import timedelta

from .utils.ut_fileimport import FileUtils
from .utils.ut import Utils


class FileImport():

    def __init__(self, core):
        global client
        client = core
        self.raiseException = client.raiseException
  
    def importNewInventory(self, filePath:str, delimiter:str):
        """
        Creates a new inventory from a CSV file
        
        Parameters:
        -----------
        filePath : str
            The file path of the csv file that should be imported.
        delimiter : str
            The CSV delimiter. Choose ',', ';', or 'tab'.

        Example:
        --------
        >>> createInventoryFromCsv(filePath='C:\\temp\\CreateMeterData.csv', delimiter=';')          
        """

        files = FileUtils._checkFilePath(filePath, self.raiseException)
        if files == None: return
        for file in files:
            content = FileUtils._readCsvFile(file, delimiter)

            ## CHECK FILE
            if content[0][0] != 'name':
                Utils._error(self, f"Wrong format. Expected header 'name' (for inventory) at position (0, 0).")
                return
    
            if content[2][0] != 'name':
                Utils._error(self, f"Wrong format. Expected header 'name' (for property) at position (2, 0).")
                return

            inventoryName = content[1][0]

            if not inventoryName:
                Utils._error(self, f"Inventory name missing")
                return 
    
            ## PREPARE IMPORT
            propertyList =[]   
            boolKeys = ['nullable', 'isArray', 'isReference'] 
            keys = [item for item in content[2]]
            columns = len(keys)

            for i, row in enumerate(content):
                if i >= 3:
                    propertyDict = {}
                    for column in range(columns):
                        if content[2][column] in boolKeys:
                            if row[column] == 'false': value = False
                            if row[column] == 'true': value = True
                        elif not row[column]: continue
                        else: value = row[column]
                        propertyDict.setdefault(content[2][column], value)
                    propertyList.append(propertyDict)

            ## IMPORT
            logger.debug(propertyList)
            result = client.createInventory(inventoryName, propertyList)
            if result == {'createInventory': {'errors': None}}: 
                logger.info(f"Inventory {inventoryName} created.")

        return

    def importItems(
            self, 
            filePath:str, 
            inventoryName:str, 
            delimiter:str=',',
            chunkSize:int = 5000, 
            pause:int = 1
            ) -> None:
        """
        Imports  items from a CSV file. The CSV file only needs a header of
        property definitions. Each line below the header represents a new item.

        Parameters:
        -----------
        filePath : str
            The file path of the csv file that should be imported.
        inventoryName : str
            The field name of the inventory.
        delimiter : str = ','
            The CSV delimiter. Choose ',', ';', or 'tab'.
        chunkSize : int = 5000
            Determines the number of items which are written per chunk. Using chunks
            can be necessary to avoid overloading. Default is 500 items per chunk.
        pause : int = 1
            Between each chunk upload a pause is inserted in order to avoid overloading.

        Example:
        --------
        >>> importItems(filePath='C:\\temp\\Items.csv', delimiter=';'
            inventoryName='meterData')          
        """
        ## TIMEZONE
        #timeZone = FileUtils._timeZone(timeZone)
               
        ## READ FILE
        files = FileUtils._checkFilePath(filePath, self.raiseException)
        if files == None: return
        for file in files:
            content = FileUtils._readCsvFile(file, delimiter)

            ## PREPARE IMPORT   
            properties = client.propertyList(inventoryName, references=True)
            logger.debug(f'Property names: {properties}')

            diff = FileUtils._comparePropertiesBasic(properties, content[0])
            if len(diff) > 0:
                Utils._error(self, f"Unknown properties: {list(diff)}")
                return 

            properties = client.inventoryProperties(inventoryName)

            referenceFields, referenceErrors = FileUtils._checkReferences(content[0])
            if len(referenceErrors) != 0:
                Utils._error(f"Multilevel references such as {referenceErrors} are not supported.")
                return

            dataType, isArray, nullable = FileUtils._analyzeProperties(inventoryName, properties)
            logger.info(f"File '{filePath}' read and properties analyzed")

            if len(referenceFields) != 0:
                referenceMapping = FileUtils._createReferenceMapping(client, inventoryName, referenceFields, isArray, content)
            else:
                referenceMapping = None

            items = FileUtils._createItems(content, dataType, isArray, nullable, referenceMapping)
            logger.debug(f'Basic items: {items}' )
            if len(items) == 0:
                logger.info(f"The file '{file}' did not contain any valid items to import.")
                continue

            # ## IMPORT
            if len(items) > chunkSize:
                lenResult = 0
                for i in range(0, len(items), chunkSize):
                    result = client.addItems(inventoryName, items[i : i + chunkSize])
                    logger.info(f"{len(result)+lenResult} items of {len(items)} imported.")
                    sleep(pause)
            else:
                result = client.addItems(inventoryName, items)
                logger.info(f"{len(result)} items of file '{filePath}' imported.")

        return

    def importTimeSeriesItems(
        self, 
        filePath:str,  
        inventoryName:str, 
        delimiter:str=',',
        chunkSize:int = 500, 
        pause:int = 1
        ) -> None:
        """
        Imports time series inventory items from a CSV file. The CSV file only needs a header of
        property definitions. Each line below the header represents a new time series.

        Parameters:
        -----------
        filePath : str
            The file path of the csv file that should be imported.
        inventoryName : str
            The field name of the inventory.
        delimiter : str = ','
            The CSV delimiter. Choose ',', ';', or 'tab'.
        chunkSize : int = 500
            Determines the number of items which are written per chunk. Using chunks
            can be necessary to avoid overloading. Default is 50 items per chunk.
        pause : int = 1
            Between each chunk upload a pause is inserted in order to avoid overloading.
                 
        """

        # if timeZone == None:
        #     timeZone = core._getDefaults()['timeZone']
               
        ## READ FILE
        files = FileUtils._checkFilePath(filePath, self.raiseException)
        if files == None: return
        for file in files:
            content = FileUtils._readCsvFile(file, delimiter)

            ## PREPARE IMPORT
            tsProperties = ['unit', 'timeUnit', 'factor']
            for header in tsProperties:
                if not header in content[0]:
                    Utils._error(self, f"Header {header} not found. Import aborted.")
                    return 

            properties = client.propertyList(inventoryName, references=True)
            logger.debug(f'Property names: {properties}')

            diff = FileUtils._comparePropertiesTimeSeries(properties, content[0])
            if len(diff) > 0:
                Utils._error(self, f"Unknown properties: {list(diff)}")
                return 

            properties = client.inventoryProperties(inventoryName)

            referenceFields, referenceErrors = FileUtils._checkReferences(content[0])

            if len(referenceErrors) != 0:
                Utils._error(f"Multilevel references such as {referenceErrors} are not supported.")
                return

            dataType, isArray, nullable = FileUtils._analyzeProperties(inventoryName, properties)
            logger.info(f"File '{filePath}' read and properties analyzed")

            if len(referenceFields) != 0:
                referenceMapping = FileUtils._createReferenceMapping(client, inventoryName, referenceFields, isArray, content)
            else:
                referenceMapping = None

            #timeSeriesItems = FileUtils._createTimeSeriesItems(content, dataType, isArray, nullable)
            timeSeriesItems = FileUtils._createItems(content, dataType, isArray, nullable, referenceMapping, type='timeSeries')
            logger.debug(f'Time series items: {timeSeriesItems}' )
            if len(timeSeriesItems) == 0:
                logger.info(f"The file '{file}' did not contain any valid items to import.")
                continue



            ## IMPORT
            if len(timeSeriesItems) > chunkSize:
                lenResult = 0
                for i in range(0, len(timeSeriesItems), chunkSize):
                    result = client.TimeSeries.addTimeSeriesItems(inventoryName, timeSeriesItems[i : i + chunkSize])
                    logger.info(f"{len(result)+lenResult} items of {len(timeSeriesItems)} imported. Waiting {pause} second(s) before continuing...")
                    sleep(pause)
            else:
                result = client.TimeSeries.addTimeSeriesItems(inventoryName, timeSeriesItems)
                logger.info(f"{len(result)} items of {len(timeSeriesItems)} imported.") 
        return

    def importTimeSeriesData(
        self, 
        filePath:str, 
        inventoryName:str, 
        importKeyProperty:str=None, 
        delimiter:str=',', 
        timeZone:str=None, 
        dateTimeFormat:str=None, 
        fromTimepoint:str=None, 
        toTimepoint:str=None, 
        timeDelta:timedelta=None, 
        chunkSize:int=20000
        ) -> None: 
        """
        Imports time series data from a specific CSV file or a folder with multiple
        CSV files. The first column is the timestamp index, whereas the first row
        consists of inventory item ids or an import key from the time series property
        definitions. Time series values are spanned as matrix between first column and
        first row. 

        Parameters:
        -----------
        filePath: str
            A path to a folder or a specific CSV file
        inventoryName: str
            The field name of the inventory, if not provided in the import file. 
            In the import file the inventory nameis located at position (0,0).
        importKeyProperty: str = None
            By default the inventory item id is used to find map columns with 
            values with time series. As an alternative, the content of a specific
            property can be used as header which will be mapped with the time series.
            This property should be unique.
        delimiter: str = ','
            The CSV delimiter. Choose ',', ';', or 'tab'.
        timeZone: str = None
            A time zone provided in IANA or isoformat (e.g. 'Europe/Berlin' or 'CET'). Defaults
            to the local time zone.
        dateTimeFormat: str = None
            Several date-time formats are supported, however, a custom format according to
            datetime.strftime() and strptime() format codes can be passed to convert the
            timestamp.
        fromTimepoint: str = None
            Specify a timestamp in isoformat from which data should be imported.
        toTimepoint: str = None
            Specify a timestamp in isoformat until which data should be imported.
        timeDelta: datetime.timedelta = None
            Define a time delta to add or substract to the original timestamp.
        chunkSize: int = 20000
            Determines the number of time series datapoints which are written per chunk. Using chunks
            can be necessary to avoid overloading.

        Example:
        --------
        >>> client.FileImport.importTimeSeriesData(
            filePath=file,
            inventoryName='meterData'
            importKeyProperty='name'
            delimiter=';', 
            timeZone='CET', 
            dateTimeFormat='%Y-%b-%d %H:%M:%S'
            fromTimePoint='2023-01-01'
            toTimepoint='2023-03-05'
            timeDelta=-timedelta(hours=1))
            chunkSize=20000
        """

        ## TIMEZONE
        timeZone = FileUtils._timeZone(timeZone)

        ## CHECK FILE PATH
        files = FileUtils._checkFilePath(filePath, self.raiseException)
        if files == None: return

        output = {}
        
        for file in files:

            fileName = file.name
        
            output.setdefault(fileName, 
                {'Time series': 0,
                'Time series errors': 0,
                'Values written': 0,
                'Value errors': 0,
                'Errors': None})

            ## READ FILE
            content = FileUtils._readCsvFile(file, delimiter)

            tsLength = len(content) - 1

            # CONVERT DATETIME COLUMN
            if dateTimeFormat == None:
                dateTimeFormat = FileUtils._dateFormat(content[1][0])

            for i, row in enumerate(content):
                try:
                    if content[i][0] == '': continue
                    content[i][0] = FileUtils._convertTimestamp(content[i][0], timeZone, dateTimeFormat, timeDelta)
                except Exception as err: 
                    if i >= 1:
                        Utils._error(self, f"Timestamp {row[0]} could not be converted. {err}")
                        return 
                    pass

            ## GET ITEM ID FROM IMPORT KEY
            if importKeyProperty != None:
                try:
                    items = client.items(inventoryName, fields=['unit', 'resolution', 'sys_inventoryItemId', importKeyProperty])
                except Exception as err:
                    Utils._error(self, err)
                    return 
                names = content[0].copy()
                del names[0]
                items = items[items[importKeyProperty].isin(names)]
                idMapping = {}
                for item in items.iterrows():
                    idMapping.setdefault(item[1][importKeyProperty],item[1]['sys_inventoryItemId'])

                if len(idMapping) == 0:
                    Utils._error(self, err)
                    return 

                logger.debug(f"Id Mapping: {idMapping}")
            else:
                try:
                    items = client.items(inventoryName, fields=['unit', 'resolution', 'sys_inventoryItemId'])
                    logger.debug("Inventory read for default option (import with ivnentory item ids).")
                except Exception as err:
                    Utils._error(self, err)
                    return 

            # Get the Inventory Id
            try:                      
                inv = client.inventories(where=f'name eq "{inventoryName}"')
                inventoryId = inv.loc[0, 'inventoryId']
                logger.debug(f"Found inventoryId {inventoryId} for inventory {inventoryName}.")
            except:
                Utils._error(self, err)
                return 

            ## VERIFY IDS, CREATE DATA_DICTS, IMPORT
            errorDict = {}
            inv = None
            tsItems = [] # only used in bulk operation

            for column in range(1, len(content[0])):
                if importKeyProperty != None:
                    try:
                        inventoryItemId = idMapping[content[0][column]]
                    except:
                        logger.warning(f"ImportKeyProperty '{content[0][column]}' not found.")
                        output[fileName]['Time series errors'] += 1
                        errorDict.setdefault(content[0][column], 'Not found')
                        continue
                else:
                    try:
                        inventoryItemId = content[0][column]
                    except:
                        logger.warning(f"Inventory item id {inventoryItemId} not found.")
                        errorDict.setdefault(content[0][column], 'Not found')
                        continue                

                try:
                    properties = items[items['sys_inventoryItemId'] == inventoryItemId]
                    if properties.empty:
                        logger.warning(f"Inventory item id {inventoryItemId} not found.")
                        output[fileName]['Time series errors'] += 1
                        errorDict.setdefault(content[0][column], 'Not found')
                        continue
                except:
                    logger.warning(f"Inventory item id {inventoryItemId} not found.")
                    continue

                properties = properties.to_dict('records')[0]
                timeUnit = properties['resolution'].split(' ')[-1]
                factor = properties['resolution'].split(' ')[0]

                tsDict = {
                        'sys_inventoryId': inventoryId,
                        'sys_inventoryItemId': None,
                        'data': {
                            'resolution': {
                                'timeUnit': None,
                                'factor': None
                            },
                            'unit': None,
                            'dataPoints': None
                        }
                    }

                valueList = []
                for i, row in enumerate(content):
                    if i >= 1:     
               
                        try:
                            if row[0] == '': continue
                            if fromTimepoint:
                                if row[0] < fromTimepoint: continue
                            if toTimepoint:
                                if row[0] > toTimepoint: continue
                            float(row[column])              
                            valueList.append({'timestamp': row[0], 'value': row[column]})
                        except:
                            errorDict.setdefault(content[0][column], {})
                            errorDict[content[0][column]].setdefault(row[0], row[column])
                            output[fileName]['Value errors'] += 1     

                logger.debug(f"Value list first 5 items: {valueList[:5]}")

                tsDict['sys_inventoryItemId'] = inventoryItemId
                tsDict['data']['unit'] = properties['unit']
                tsDict['data']['resolution']['timeUnit'] = timeUnit
                tsDict['data']['resolution']['factor'] = factor  
                tsDict['data']['dataPoints'] = valueList

                tsItems.append(tsDict)
                output[fileName]['Values written'] += len(valueList)

            logger.debug("Time series collection created")
            output[fileName]['Time series'] = len(tsItems)

            tsItemsEmpty = copy.deepcopy(tsItems)
            for item in tsItemsEmpty:
                del item['data']['dataPoints']
            
            for i in range(0, tsLength, chunkSize):
                tsChunk = copy.deepcopy(tsItemsEmpty)
                for j, ts in enumerate(tsChunk):
                    tsChunk[j]['data'].setdefault('dataPoints', tsItems[j]['data']['dataPoints'][i : i + chunkSize])
                    
                client.TimeSeries.setTimeSeriesDataCollection(timeSeriesData=tsChunk)    
                logger.info(f"({int(i/chunkSize+1)}/{tsLength//chunkSize+1}) chunks imported.")            

        logger.info(f"Import finished")

        return 

    def importGroupInstanceItems(
            self, 
            filePath:str, 
            groupInventoryName:str,
            instanceInventoryName, 
            importKeyProperty, 
            delimiter:str=None, 
            chunkSize:int = 500,
            pause:int = 1
            ) -> None:
        """
        Imports  group instance items from a CSV file. The CSV file only needs a header
        of property definitions. The first column is reserved for the groupInventoryItemId.
        Each line below the header represents a new item.

        Parameters:
        -----------
        filePath: str
            The file path of the csv file that should be imported.
        groupInventoryName: str
            The field name of the group inventory.
        instanceInventoryName: str
            The field name of the time series instance inventory belonging to the group.
        importKeyProperty: str
            Provide a property of the parent group time series that is unique.
            (The first column 'groupName' used in the CSV file.)
        delimiter: str = ','
            The CSV delimiter. Choose ',', ';', or 'tab'.
        chunkSize: int = 500
            Determines the number of items which are written per chunk. Using chunks
            can be necessary to avoid overloading.
        pause: int = 1
            Between each chunk upload a pause is inserted in order to avoid overloading.

        """
              
        ## READ FILE
        files = FileUtils._checkFilePath(filePath, self.raiseException)
        if files == None: return

        for file in files:
            content = FileUtils._readCsvFile(file, delimiter)

            ## PREPARE IMPORT
            properties = client.propertyList(instanceInventoryName, references=True)
            logger.debug(f'Property names: {properties}')

            diff = FileUtils._comparePropertiesBasic(properties, content[0][1:-1])
            if len(diff) > 0:
                Utils._error(self, f"Unknown properties: {list(diff)}")
                return 

            properties = client.inventoryProperties(instanceInventoryName)

            if importKeyProperty != None:
                try:
                    items = client.items(groupInventoryName, fields=['sys_inventoryItemId', importKeyProperty])
                except Exception as err:
                    Utils._error(self, err)
                    return
            
                names = [row[0] for row in content]
                del names[0]

                items = items[items[importKeyProperty].isin(names)]

                idMapping = {}
                for item in items.iterrows():
                    idMapping.setdefault(item[1][importKeyProperty],item[1]['sys_inventoryItemId'])
                logger.debug(f"Id mapping: {idMapping}")
                if len(idMapping) == 0:
                    Utils._error(self, f"No item ids for importKeyProperty '{importKeyProperty}' found.")
                    return 

            dataType, isArray, nullable = FileUtils._analyzeProperties(instanceInventoryName, properties)
            logger.debug(f'Data types: {dataType}')
            logger.debug(f'Array properties: {isArray}')
            logger.debug(f'Nullable properties: {nullable}')
            logger.info(f"File '{filePath}' read and properties analyzed")

            items = FileUtils._createInstanceItems(content, dataType, isArray, nullable, idMapping)
            logger.debug(f'Instance items: {items}' )

            # ## IMPORT
            if len(items) > chunkSize:
                lenResult = 0
                for i in range(0, len(items), chunkSize):
                    result = client.TimeSeries.addTimeSeriesItemsToGroups(groupInventoryName, items[i : i + chunkSize])
                    logger.info(f"{len(result)+lenResult} items of {len(items)} imported.")
                    sleep(pause)
            else:
                result = client.TimeSeries.addTimeSeriesItemsToGroups(groupInventoryName, items)
                logger.info(f"{len(items)} items of file '{filePath}' imported.")

        return

    def importGroupInstanceItemsWithData(
            self, 
            filePath:str, 
            groupInventoryName:str,
            instanceInventoryName:str, 
            groupKeyProperty:str=None, 
            instanceKeyProperties:list=None,
            delimiter:str=',', 
            timeZone:str=None, 
            dateTimeFormat:str=None, 
            fromTimepoint:str=None, 
            toTimepoint:str=None, 
            timeDelta:timedelta=None, 
            chunkSize=20000, 
            pause:int = 1
            ) -> None:
        """
        Imports  group instance items from a CSV file. The CSV file only needs a header
        of property definitions. The first column is reserved for the groupInventoryItemId.
        Each line below the header represents a new item.

        Parameters:
        -----------
        filePath : str
            The file path of the csv file that should be imported.
        groupInventoryName : str
            The field name of the group inventory.
        instanceInventoryName : str
            The field name of the time series instance inventory belonging to the group.
        groupKeyProperty : str
            Is a property of the group item to identify it, which have to be placed 
            in the first line of the file. If None, the group item id is expected. 
        instanceKeyProperties:
            One or two key properties of the instance item to identiy it uniquely. The order must be
            the same as in the import file.
        delimiter : str = ','
            The CSV delimiter. Choose ',', ';', or 'tab'.
        timeZone: str = None
            A time zone provided in IANA or isoformat (e.g. 'Europe/Berlin' or 'CET'). Defaults
            to the local time zone.
        dateTimeFormat: str = None
            Several date-time formats are supported, however, a custom format according to
            datetime.strftime() and strptime() format codes can be passed to convert the
            timestamp.
        fromTimepoint : str = None
            Specify a timestamp in isoformat from which data should be imported.
        toTimepoint : str = None
            Specify a timestamp in isoformat until which data should be imported.
        timeDelta: datetime.timedelta = None
            Define a time delta to add or substract to the original timestamp.
        chunkSize: int = 20000
            Determines the number of time series datapoints which are written per chunk. Using chunks
            can be necessary to avoid overloading.
        pause: int = 1
            Between each chunk upload a pause is inserted in order to avoid overloading.
        """

        ## TIMEZONE
        timeZone = FileUtils._timeZone(timeZone)

        ## CHECK FILE PATH
        files = FileUtils._checkFilePath(filePath, self.raiseException)
        if files == None: return
    
        output = {}
        
        for file in files:

            fileName = file.name
        
            output.setdefault(fileName, 
                {'Time series instances': 0,
                'Time series instance errors': 0,
                'Values written': 0,
                'Value errors': 0,
                'Errors': None})

            ## READ FILE
            content = FileUtils._readCsvFile(file, delimiter)

            ## PREPARE IMPORT: Get Instance properties
            instanceProperties = []
            for i, row in enumerate(content):
                if i == 0: continue
                if row[0] in ['unit', 'timeUnit', 'factor']: 
                    pass
                elif row[0] == 'data':
                    dataBegin = i + 1
                    break
                elif row[0] == 'values':
                    dataBegin = i + 1
                    break
                else:
                    instanceProperties.append(row[0])
                if i > 100: 
                    Utils._error(self, f"No keyword 'data' or 'values' found")
                    return

            tsLength = len(content) - dataBegin

            # CONVERT DATETIME COLUMN
            if dateTimeFormat == None:
                dateTimeFormat = FileUtils._dateFormat(content[dataBegin][0])

            for i, row in enumerate(content[dataBegin:]):
                try:
                    if row[0] == '': continue
                    content[dataBegin+i][0] = FileUtils._convertTimestamp(content[dataBegin+i][0], timeZone, dateTimeFormat, timeDelta)
                except Exception as err: 
                    if i >= 1:
                        Utils._error(self, f"Timestamp {row[0]} could not be converted. {err}")
                        return 
                    pass

            ## PREPARE IMPORT: Compare Instance properties
            properties = client.propertyList(instanceInventoryName, references=True)
            logger.debug(f'Property names: {properties}')

            diff = FileUtils._comparePropertiesBasic(properties, instanceProperties)
            if len(diff) > 0:
                Utils._error(self, f"Unknown properties: {list(diff)}")
                return 
            
            properties = client.inventoryProperties(instanceInventoryName)

            ## PREPARE IMPORT: Get Mapping of Import Key
            if groupKeyProperty != None:
                try:
                    items = client.items(groupInventoryName, fields=['sys_inventoryItemId', groupKeyProperty])
                except Exception as err:
                    Utils._error(self, err)
                    return 
            
                names = [column for column in content[0]]
                del names[0]

                items = items[items[groupKeyProperty].isin(names)]

                idMapping = {}
                for item in items.iterrows():
                    idMapping.setdefault(item[1][groupKeyProperty],item[1]['sys_inventoryItemId'])
                logger.debug(f"Id mapping: {idMapping}")
                if len(idMapping) == 0:
                    Utils._error(self, f"No item ids for groupKeyProperty '{groupKeyProperty}' found.")
                    return 

            ## PREPARE IMPORT: Check dataType, Array and nullable properties
            dataType, isArray, nullable = FileUtils._analyzeProperties(instanceInventoryName, properties)
            logger.info(f"File '{filePath}' read and properties analyzed")

            itemContent = FileUtils._createInstanceItemContent(content[:dataBegin])
            tsItemContent = content[:dataBegin]
            items = FileUtils._createInstanceItems(itemContent, dataType, isArray, nullable, idMapping, transpose=True)
            tsItems = FileUtils._createInstanceItems(tsItemContent, dataType, isArray, nullable, idMapping, transpose=True)

            logger.debug(f'Instance items: {items}' )

            # Find position of instance key properties
            if instanceKeyProperties == None:
                Utils._error(self, f"No instanceKeyProperties provided.")
                return 

            instancePropPos = []
            for i, row in enumerate(content[:dataBegin]):
                for property in instanceKeyProperties:
                    if property == row[0]:
                        instancePropPos.append(i)

            try:                      
                inv = client.inventories(where=f'name eq "{instanceInventoryName}"')
                inventoryId = inv.loc[0, 'inventoryId']
                logger.debug(f"Found inventoryId {inventoryId} for inventory {instanceInventoryName}.")
            except:
                Utils._error(self, f"No inventory with name '{instanceInventoryName}'.")
                return 

            ## IMPORT: create group instance items
            errorDict = {}

            for i in range(0, len(items)):
                try:
                    client.TimeSeries.addTimeSeriesItemsToGroups(groupInventoryName, [items[i]])
                    sleep(pause)
                except:
                    logger.warning(f"Instance item in column {i+1} could not be created")
                    errorDict.setdefault(i+1, "Instance could not be created")
                try:
                    if len(instanceKeyProperties) == 1:
                        x0 = content[instancePropPos[0]][i+1]
                        instanceTs = client.items(instanceInventoryName, 
                            where=f'{instanceKeyProperties[0]} eq "{x0}"')
                    if len(instanceKeyProperties) >= 2:
                        x0 = content[instancePropPos[0]][i+1]
                        x1 = content[instancePropPos[1]][i+1]
                        instanceTs = client.items(instanceInventoryName, 
                            where=f'{instanceKeyProperties[0]} eq "{x0}" and {instanceKeyProperties[1]} eq "{x1}"')
                    inventoryItemId = instanceTs.loc[0, 'sys_inventoryItemId']
                except:
                    logger.warning(f"Instance item in column {i+1} not found")
                    output[fileName]['Time series instance errors'] += 1
                    continue
                sleep(pause)

                tsDict = {
                    'sys_inventoryId': inventoryId,
                    'sys_inventoryItemId': inventoryItemId ,
                    'data': {
                        'resolution': {
                            'timeUnit': tsItems[i]['timeUnit'],
                            'factor': tsItems[i]['factor']
                        },
                        'unit': tsItems[i]['unit'],
                        'dataPoints': None
                    }
                }

                valueList = []
                for row in content[dataBegin:]:
                    try:
                        if fromTimepoint:
                            if row[0] < fromTimepoint: continue
                        if toTimepoint:
                            if row[0] > toTimepoint: continue
                        float(row[i+1])   
                        valueList.append({'timestamp': row[0], 'value': row[i+1]})
                    except:
                        pass
                        errorDict[i+1] = {}
                        errorDict[i+1].setdefault(row[0], row[i+1])
                        output[fileName]['Value errors'] += 1
               
                for k in range(0, tsLength, chunkSize):     
                    tsDict['data']['dataPoints'] = valueList[k : k + chunkSize]
                    try:
                        client.TimeSeries.setTimeSeriesDataCollection(timeSeriesData=[tsDict])
                    except Exception as err:
                        logger.error(f"Time series values could not be written. Cause: {err}")
                        errorDict[i+1] = err
                        break

                output[fileName]['Values written'] += len(valueList)
            if len(errorDict) > 0:
                output[fileName]['Errors'] = errorDict
                        
        return output

    def importReferences(
        self, 
        filePath:str, 
        delimiter:str=','
        ) -> dict:
        """
        Connects reference fields of existing items of two inventories. 
        Connect array-type [n:1] and non array references [1:1].
        Use FileImport.ImportReferences() to generate a csv template.

        Parameters:
        -----------
        filePath : str
            The file path of the csv file that should be imported.
        delimiter : str = ','
            The CSV delimiter. Choose ',', ';', or 'tab'.

        Example:
        --------
        >>> filePath = r'C:\temp\importReferences.csv'
        >>> client.FileImport.importReferences(filePath, delimiter=';')
        """

        def _readInventory(
                inventoryName:str, 
                property:str,
                items:list) -> pd.DataFrame:

            filterList = '['
            for item in items:
                filterList += f'"{item}",' 
            filterList += ']' 
            
            df = client.items(inventoryName, fields=[property,'sys_inventoryItemId'], 
                where=f'{property} in {filterList}')
            df.set_index(property, inplace=True)
            logger.debug(f"Items to reference (first 5):\n {df.iloc[:5]}")
            return df

        def _mapItems(
                content:list, 
                refItems:pd.DataFrame, 
                array:bool=True) -> dict:

            data = {}
            for i, row in enumerate(content):
                if i <= 1: continue
                if row[0] == '':
                    msg = f"No value in line {i}"
                    logger.warning(msg)
                    errors.setdefault(f'Row {i}', 'No value')
                    continue
                try:
                    if array == True:
                        data.setdefault(row[0], [])
                        data[row[0]].append(refItems['sys_inventoryItemId'][row[1]])
                    else:
                        data.setdefault(row[0], refItems['sys_inventoryItemId'][row[1]])
                except KeyError as err:
                    msg = f"Row {i}: unknown item {row[0]}"
                    logger.warning(msg)
                    errors.setdefault(f'Row {i}', f'unknown reference item {row[1]}')
                    continue
                except Exception as err:
                    msg = f"Row {i}: unknown error: {row[1]}"
                    logger.warning(msg)
                    errors.setdefault(f'Row {i}', f'unknown error: {row[1]}')
                    continue
            return data

        files = FileUtils._checkFilePath(filePath, self.raiseException)
        if files == None: return

        inventories = client.inventories(fields=['name', 'inventoryId'])

        output = {}
        
        for file in files:

            fileName = file.name
            errors = {}

            output.setdefault(fileName, 
                {'Items updated': 0,
                'Errors': None})
            content = FileUtils._readCsvFile(file, delimiter)
            
            # Read reference field
            try:
                inventory = content[0][0].split('.')[0]
                referenceField = content[0][0].split('.')[1]
            except:
                msg = f"'{content[0][0]}' is not a valid 'inventory.property' format."
                logger.error(f"{fileName}: {msg}")
                output[fileName]['Errors'] = msg
                continue

            # Arrange data
            fields = [column for column in content[1]]
            logger.debug(f"Fields: {fields}")
            invFields = {}
            inventoryList = []
            for item in fields:
                try:
                    splitItem = item.split('.')
                    invFields.setdefault(splitItem[0], splitItem[1])
                    inventoryList.append(splitItem[0])
                except:
                    msg = f"'{item}' is not a valid 'inventory.property' format."
                    logger.error(f"{fileName}: {msg}")
                    output[fileName]['Errors'] = msg
                    continue

            # Check if inventory is in inventories
            inventoryFieldsCorrect = True
            for key in invFields.keys():
                if key not in list(inventories['name']):
                    inventoryFieldsCorrect = False
                    msg = f"'{key}' is not a known inventory"
                    logger.error(f"{fileName}: {msg}")
                    output[fileName]['Errors'] = msg
                    continue                 
            if inventoryFieldsCorrect == False: return
            logger.debug(f"Inventories: {inventoryList}")

            # Check type of first field
            properties = client.inventoryProperties(inventory)
            properties.set_index('name', inplace=True)
            if properties['type'][referenceField] != 'reference':
                msg = f"'{property}' is not a reference field"
                logger.error(f"{fileName}: {msg}")
                output[fileName]['Errors'] = msg
                continue       

            if properties['isArray'][referenceField] == True:

                referenceValues = [row[1] for i, row in enumerate(content) if i > 1]
                logger.debug(f"Reference Values (first 5): {referenceValues[:5]}")

                # Read inventory items to reference    
                dfReferenceItems = _readInventory(inventoryList[1], invFields[inventoryList[1]], referenceValues)

                # Create Data Dict
                mappedItems = _mapItems(content, dfReferenceItems, array=True)
                   
                # Read inventory items to update
                dfUpdateItems = _readInventory(inventory, invFields[inventory], mappedItems.keys())

                # Create connections
                itemsUpdated = 0
                itemsReferred = 0
                for key, value in mappedItems.items():
                    try:
                        client.updateArrayProperty(inventory, dfUpdateItems['sys_inventoryItemId'][key], referenceField, 'insert', value)
                        if len(value) != 0:
                            itemsUpdated += 1
                        itemsReferred += len(value)
                    except Exception as err:
                        msg = f"Item: {dfUpdateItems['sys_inventoryItemId'][key]}, Value: {value}, Problem: {err}"
                        logger.warning(err)
                        errors.setdefault({dfUpdateItems['sys_inventoryItemId'][key]}, [f'Value: {value}', f'Problem: {err}'])

                output[fileName]['Items updated'] = itemsUpdated
                output[fileName].setdefault('Items referred', itemsReferred)
                if len(errors) != 0:
                    output[fileName]['Errors'] = errors

                return output

            else: #If it is not an array (1:1 reference)
                referenceValues = [row[1] for i, row in enumerate(content) if i > 1]
                logger.debug(f"Reference Values (first 5): {referenceValues[:5]}")

                dfReferenceItems = _readInventory(inventoryList[1], invFields[inventoryList[1]], referenceValues)

                # Create Data Dict
                mappedItems = _mapItems(content, dfReferenceItems, array=False)
                    
                # Read inventory items to update
                dfUpdateItems = _readInventory(inventory, invFields[inventory], mappedItems.keys())

                # Create connections
                updateItems = []
                for key, value in mappedItems.items():
                    item = {
                        'sys_inventoryItemId': dfUpdateItems['sys_inventoryItemId'][key],
                        referenceField: value
                    }
                    updateItems.append(item)
                client.updateItems(inventory, updateItems)

                output[fileName]['Items updated'] = len(updateItems)
                if len(errors) != 0:
                    output[fileName]['Errors'] = errors

                return output

    def templateImportReferences(
        self, 
        filePath:str=None
        ) -> pd.DataFrame:
        """Provides a template to import item references."""
        content = [
            ['updateInventory.propertyToBeUpdated', '', '// Enter inventory and property to be updated, like "marketData.qualityCriteria"'],
            ['updateInventory.uniqueProperty', 'referenceInventory.uniquePropertyOfReference', '// Enter inventory name to be updated and a unique property like "marketData.name" and a referring inventory with another unique property like "qualityCriteria.name"'],
            ['Swissgrid AE CH long', 'Swissgrid AE CH long', '// Example data, repeat entries and use same item in first column to import array items'],
            ['Swissgrid AE CH short', 'Swissgrid AE CH short','// Example data, repeat entries and use same item in first column to import array items']
        ]
        header = ['column 1', 'column 2', 'Remarks']

        df = pd.DataFrame(content, columns=header)
        if filePath != None:
            df.to_csv(filePath, index=False, header=False)
        return df
       

            


from gql import Client
from gql.transport.requests import RequestsHTTPTransport
from loguru import logger
import base64

import pandas as pd
import sys
import json
import os
import requests
import getpass

from .fileimport import FileImport
from .automation import Automation, Schedule
from .programming import Programming

from .utils.ut import Utils, Structure, Defaults
from .timeseries import TimeSeries

import warnings

class TechStack():
    """
    Initializes a Seven2one TechStack client.

    Parameters:
    ----------
    host: str
        The host domain or ip address, e.g. 'app.organisation.com'
    user: str
        A username
    password: str = None
        The password belonging to the user. If None, it needs to
        be entered interactively.
    accessToken: str = None
        The access token. Can be used alternatively instead of user and password.
    proxies: dict = None
        Provide one or more proxy addresses, e.g.: {'https':'159.98.7.123'} 
    logLevel: str = 'WARNING'
        Change the highest log level. Available log levels are 
        'ERROR', 'WARNING', 'INFO' and 'DEBUG'.
    usePorts: bool = False
        A developer feature: if True, ports with prefix '8' for a
        developer ComposeStack will be used.
    timeZone: str = None
        A time zone provided in IANA or isoformat (e.g. 'Europe/Berlin' or 'CET').
        Defaults to the local timezone. 
    dateTimeOffset: bool = True
        Choose, if a Timestamp should be displayed as time zone naive (dateTimeOffset = False) 
        or with Offset information. 
    raiseException: bool = False
        In default mode (False) exceptions will be avoided to provide better user experience.
        True is recommended in automated processes to avoid silent failures.
    copyGraphQLString: bool = False
        A developer feature: if True, with each method execution, 
        the graphQL string is copied to the clipboard (Windows only).

    Examples:
    >>> client = TechStack('app.orga.com/', 'admin', 'xxxx', logLevel='INFO')
    """

    def __init__(
        self,
        host: str,
        user: str = None,
        password: str = None,
        accessToken: str = None,
        proxies: dict = None,
        logLevel: str = 'WARNING',
        usePorts: bool = False,
        timeZone: str = None,
        dateTimeOffset: bool = True,
        raiseException: bool = False,
        copyGraphQLString: bool = False,
        includeSystemProperties: bool = False,
        timeout: int = 30
        ) -> object:

        warnings.filterwarnings('ignore')
             
        self.raiseException = raiseException

        try:
            logger.remove()

            if logLevel in ['DEBUG', 'TRACE']:
                logger_format = "{level:<10} {time} {module}.{function} {line}: {message}"
                logger_diagnose = True
                logger_backtrace = True
            else:
                logger_format = "{level:<10} {time} {message}"
                logger_diagnose = False
                logger_backtrace = False

            def stdout_log_filter(record):
                return record["level"].no < logger.level('ERROR').no

            logger.add(sys.stdout, format=logger_format, level=logLevel, catch=True, diagnose=logger_diagnose, backtrace=logger_backtrace, filter=stdout_log_filter)
            logger.add(sys.stderr, format=logger_format, level='ERROR', catch=True, diagnose=logger_diagnose, backtrace=logger_backtrace)
        except:
            pass

        if os.name == 'nt': 
            logger.debug('Dectected Windows, enabling pyperclip')
        else:
            logger.debug(f"Dectected platform: {os.name}")

        if (user == None) and (accessToken == None):
            raise ValueError("User or accessToken is mandatory.")

        if (password == None) and (accessToken == None):
            password = getpass.getpass('Enter password: ')

        if usePorts == False:
            tokenUrl = f'https://{host}/authn/connect/token'
            dynEndpoint = f'https://{host}/dynamic-objects/graphql/'
            automationEndpoint = f'https://{host}/automation/graphql/'
            scheduleEndpoint = f'https://{host}/schedule/graphql/'
            programmingEndpoint = f'https://{host}/programming/graphql/'
            tsGatewayEndpoint = f'https://{host}/timeseries/graphql/'
        else: 
            tokenUrl = f'http://{host}:8040/connect/token'
            dynEndpoint = f'http://{host}:8050/graphql/'
            automationEndpoint = f'http://{host}:8120/graphql/'
            scheduleEndpoint = f'http://{host}:8130/graphql/'
            programmingEndpoint = f'http://{host}:8140/graphql/'
            tsGatewayEndpoint = f'http://{host}:8195/graphql/'

        token = accessToken if accessToken != None else self._getAccessToken(tokenUrl, user, password, proxies, timeout)
        if token == None:
            return
        logger.debug(f"Got Token: {token[:10]}...")

        header = {
            'authorization': 'Bearer ' + token
        }

        self.transport =  RequestsHTTPTransport(url=dynEndpoint, headers=header, verify=True, proxies=proxies)
        self.client = Client(transport=self.transport, fetch_schema_from_transport=True)  
        self.accessToken = token
        self.scheme = self.client.introspection
        self.proxies = proxies

        # Defaults:
        if timeZone == None:
            timeZone = Utils._timeZone('local')

        self.defaults = Defaults(
            useDateTimeOffset=dateTimeOffset, 
            copyGraphQLString=copyGraphQLString,
            includeSystemProperties=includeSystemProperties,
            timeZone=timeZone
            )

        #Get scheme
        if self.scheme == None:
            graphQLString = Structure._introspectionQueryString()
            self.scheme = Utils._executeGraphQL(self, graphQLString)

        # Structures
        self.structure = Utils._executeGraphQL(self, Structure.queryStructure)
        self.structure = Structure._fullStructureDict(self.structure)
        self.objects = Structure._fullStructureNT(self.structure)
        self.inventory = Structure._inventoryNT(self.structure)
        self.inventoryProperty = Structure._inventoryPropertyNT(self.structure)

        # Initialize further gateways
        try:
            self.TimeSeries = TimeSeries(self.accessToken, tsGatewayEndpoint, self)
            
        except Exception as err:
            logger.warning(f"Time series gateway not available")
            logger.debug(f"Reason: {err}")
            
        try:
            self.Automation = Automation(self.accessToken, automationEndpoint, self)
        except Exception as err:
            logger.warning(f"Automation service not available")
            logger.debug(f"Reason: {err}")

        try:
            self.Schedule = Schedule(self.accessToken, scheduleEndpoint, self)
        except Exception as err:
            logger.warning(f"Schedule service not available")
            logger.debug(f"Reason: {err}")

        try:
            self.Programming = Programming(self.accessToken, programmingEndpoint, self)
        except Exception as err:
            logger.warning(f"Programming service not available")
            logger.debug(f"Reason: {err}")

        self.FileImport = FileImport(self)
        return

    def _getAccessToken(self, tokenUrl: str, user: str, password: str, proxies: dict, timeout: int) -> str:
        data = {
            'client_id': 'techstack-ui',
            'grant_type': 'password',
            'username': user,
            'password': password,
            'scope': 'dynamicobjectservice'
        }

        response = requests.post(tokenUrl, data, verify=True, proxies=proxies, timeout=timeout)
        responseJson = json.loads(response.text)
        if not response.status_code == 200:
            Utils._error(self, responseJson)
            return None

        token = responseJson['access_token']
        return token

    def updateClient(self) -> None:
        """
        Updates the client scheme and structures, e.g. after adding inventories
        or new inventory properties.
        """

        self.client = Client(transport=self.transport, fetch_schema_from_transport=True)  
        self.structure = Utils._executeGraphQL(self, Structure.queryStructure)
        self.structure = Structure._fullStructureDict(self.structure)
        self.objects = Structure._fullStructureNT(self.structure)
        self.inventory = Structure._inventoryNT(self.structure)
        self.inventoryProperty = Structure._inventoryPropertyNT(self.structure)

        return

    def inventories(
        self, 
        fields:list=None, 
        where:str=None, 
        orderBy:str=None, 
        asc:bool=True
        ) -> pd.DataFrame:
        """
        Returns a DataFrame of existing inventories.

        Parameters:
        ----------
        fields: list=None
            The list of fields to be queried, e.g.
            ['name', 'inventoryId, variant.name']
        where: str=None
            Use a string to add where criteria like 'name eq "meterData"'.
            The argument is momentarily very limited.
        orderBy: str=None
            Select one field to sort by.
        asc: bool=True
            Determines the sort order of items. Set to False to apply descending order.

        Examples:
        >>> inventories()
        >>> inventories(fields=['name', 'inventoryId'], 
                where='city eq "Hamburg"', 
                orderBy='variant', asc=True)
        """

        if fields == None:
            fields = ['name', 'inventoryId', 'variant.name', 'historyEnabled', 'hasValidityPeriods', 'isDomainUserType']
            _fields = Utils._queryFields(fields, recursive=True)
        else:
            if type(fields) != list:
                fields = [fields]
            try:
                _fields = Utils._queryFields(fields, recursive=True)
            except:
                Utils._error(self, "Fields must be provided as list, e.g. ['name', 'inventoryId, variant.name']")
                return


        topLevelWhere, _ = Utils._handleWhere(self, where)

        if orderBy != None:
            if asc != True:
                _orderBy = f'order: {{ {orderBy}: DESC }}'
            else:
                _orderBy = f'order: {{ {orderBy}: ASC }}'
        else: _orderBy = ''

        graphQLString= f'''query inventories {{
        inventories 
            (pageSize:1000 {_orderBy} {topLevelWhere})
            {{
            {_fields}
            }}
        }}
        '''
        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        df = pd.json_normalize(result['inventories'])
        return df

    def items(
        self, 
        inventoryName:str, 
        references:bool=False, 
        fields:list=None,
        where:str=None, 
        orderBy:str=None, 
        asc:bool=True, 
        pageSize:int=5000, 
        arrayPageSize:int=50,
        top:int=100000,
        ) -> pd.DataFrame:
        """
        Returns items of an inventory in a DataFrame.

        Parameters:
        -----------
        inventoryName : str
            The name of the inventory.
        references: bool = False
            If True, items of referenced inventories will be added to the DataFrame. If
            the fields-parameter is used, this parameter is ignored.
        fields: list | str = None
            A list of all properties to be queried. If None, all properties will be queried.
            For referenced items use a '.' between inventory name and property.
        where: str = None
            Use a string to add where criteria like
            'method eq "average" and location contains "Berlin"'.
            Referenced items are not supported.
        orderBy: str|list = None
            Select one or more fields to sort by.
        asc: bool|list = True
            Determines the sort order of items. If set to False, a descending order 
            is applied.
        pageSize: int = 5000
            The page size of items that is used to retrieve a large number of items.
        arrayPageSize: int = 50
            The page size of list items that is used to page list items in an inventory item.
        top: int = None
            Returns a restricted set of items oriented at the latest entries.

        Example:
        --------
        >>> items('appartments', references=True)
        """

        if inventoryName not in self.inventory:
            Utils._error(self, f"'{inventoryName}' does not exist.")
            return

        # where
        try:
            topLevelWhere, resolvedFilterDict = Utils._handleWhere(self, where, inventoryName)
        except:
            return

        # core
        deleteId = False
        propertyDict = Utils._properties(self.scheme, inventoryName, recursive=True, 
            sysProperties=self.defaults.includeSystemProperties)            
        if fields != None:
            if type(fields) != list:
                fields = [fields]
            if 'sys_inventoryItemId' not in fields:
                deleteId = True
                fields += ['sys_inventoryItemId']
            _fields = Utils._queryFields(fields, propertyDict['arrayTypeFields'], arrayPageSize, 
                filter=resolvedFilterDict, recursive=True)
        else: 
            properties = Utils._propertyList(propertyDict['properties'], recursive=references)
            _fields = Utils._queryFields(properties, propertyDict['arrayTypeFields'], 
                arrayPageSize, filter=resolvedFilterDict, recursive=references)
        logger.debug(f"Fields: {_fields}")

        if len(_fields) == 0:
            Utils._error(self, f"Inventory '{inventoryName}' not found.")
            return

        result = []
        count = 0
        stop = False
        lastId = ''

        while True:

            # Handling top (premature stop)
            if top != None:
                loadedItems = pageSize * count
                if top - loadedItems < pageSize:
                    stop = True
                    pageSize = top - loadedItems

            graphQLString= f''' query getItems {{
                    {inventoryName} (
                            pageSize: {pageSize} 
                            {lastId}
                            {topLevelWhere}
                            ) {{
                        {_fields}
                    }}
                }}
                '''

            _result = Utils._executeGraphQL(self, graphQLString)     
            
            if _result[inventoryName]:
                result += _result[inventoryName]
                count += 1
            try:
                cursor = _result[inventoryName][-1]['sys_inventoryItemId']
                lastId = f'lastId: "{cursor}"'
            except: 
                break

            if stop == True: break

        df = pd.json_normalize(result)

        if orderBy != None:
            try:
                df.sort_values(orderBy, ascending=asc, inplace=True, ignore_index=True)
            except Exception as err:
                Utils._error(self, err)          
                return
        
        if fields != None: #sorts dataframe according to given fields
            try: # for array field this does not work
                df = df[fields]
            except: pass
        if deleteId: 
            try: del df['sys_inventoryItemId']
            except: pass

        return df

    def inventoryProperties(
        self, 
        inventoryName, 
        namesOnly=False
        ) -> pd.DataFrame:
        """
        Returns a DataFrame of a query of properties of an inventory.

        Parameters:
        ----------
        inventoryName : str
            The name of the inventory.
        namesOnly : bool
            If True, only property names will be returned

        Example:
        --------
        >>> inventoryProperties('appartments') 


        """

        propertyFields = f'''
            name          
            ... Scalar
            isArray
            nullable
            ... Reference
            type
            propertyId
        '''

        graphQLString= f'''query Inventory {{
        inventory
            (inventoryName: "{inventoryName}")
            {{
            properties {{
                {propertyFields}
                }}
            }}
        }}
        fragment Scalar on IScalarProperty {{
            dataType
            }}
        fragment Reference on IReferenceProperty {{
            inventoryId
            inventoryName
        }}
        '''
        
        result = Utils._executeGraphQL(self, graphQLString)
        if result['inventory'] == None:
            Utils._error(self, f"Inventory '{inventoryName}' not found.")  
            return

        df = pd.json_normalize(result['inventory']['properties'])

        if namesOnly == True:
            return list(df['name'])
        else:
            return df

    def propertyList(self, inventoryName, references=False, dataTypes=False):
        """
        Returns a list of properties of an inventory and its referenced inventories
        by reading out the scheme.

        Parameters:
        ----------
        inventoryName : str
            The name of the inventory.
        references : bool
            If True, properties of referenced inventories included.
        dataTypes : bool
            If True, result will be displayed as Series with properties as index and
            dataTypes as values.

        Example:
        --------
        >>> propertyList('appartments') 

        """

        propertyDict = Utils._properties(self.scheme, inventoryName=inventoryName, recursive=references)

        if dataTypes == False:
            properties = Utils._propertyList(propertyDict['properties'], recursive=references)
        else:
            properties = pd.Series(Utils._propertyTypes(propertyDict['properties']))

        return properties

    def filterValues(
        self, 
        inventoryName:str, 
        top:int=10000
        ) -> pd.DataFrame:
        """
        Returns a DataFrame of values that can be used in a where-string. 
        Only string data types are returned. 

        Parameters:
        ----------
        inventoryName : str
            The name of the inventory.
        top: int = None
            Uses a restricted set of items oriented at the latest entries to 
            provide a value set.

        Example:
        --------
        >>> filterValues('appartments') 

        """
        properties = self.propertyList(inventoryName, dataTypes=True)
        logger.debug(f"Properties: {properties}")

        propertyList = []
        for property, dataType in zip(properties.index, properties):
            if 'sys_' in property: continue
            if dataType != 'String': continue
            propertyList.append(property)
        logger.debug(f"PropertyList: {propertyList}")

        df = self.items(inventoryName, fields=propertyList, top=top)
        logger.debug(f"Used columns: {df.columns}")

        propertyValues = {}
        for property in propertyList:
            if 'sys_' in property: continue
            if len(set(df[property])) == len(df): continue
            propertySet = set(df[property])
            logger.debug(f"PropertySet: {propertySet}")
            try:
                propertySet.remove(None)
            except: pass
            propertyValues.setdefault(df[property].name, sorted(list(propertySet)))

        valuesDf = pd.DataFrame.from_dict(propertyValues, orient='index').T
        valuesDf.fillna(value='-', inplace=True)
        return valuesDf

    def addItems(
        self, 
        inventoryName:str, 
        items:list
        ) -> str:
        """
        Adds from a list of dicts new items and returns a list
        of inventoryItemIds.

        Parameters:
        -----------
        inventoryName : str
            The name of the inventory.
        items : list
            A list with dictionaries for each item.

        Example:
        --------
        >>> items = [
                {
                'meterId': '86IEDD99',
                'dateTime': '2020-01-01T05:50:59Z'
                },
                {
                'meterId': '45IXZ52',
                'dateTime': '2020-01-07T15:41:14Z'
                }
            ]
        >>> addItems('meterData', items)
        """
        items = Utils._propertiesToString(items)
        key = f'create{inventoryName}'

        graphQLString= f'''mutation addItems {{
            {key} (input: 
                {items}
            )
                {{
                    {Utils.errors}           
                inventoryItems {{
                    sys_inventoryItemId
                }}
            }}
        }} 
        '''
        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        if result[key]['errors']:
            Utils._listGraphQlErrors(result, key)
        return result[key]['inventoryItems']

    def updateItems(
        self, 
        inventoryName:str, 
        items:list
        ) -> None:
        """
        Updates from a list of dicts existing items. The 'sys_inventoryItemId'
        must be passed to each item.

        Parameters:
        -----------
        inventoryName : str
            The name of the inventory.
        items : list
            A list with dictionaries for each item.

        Example:
        --------
        >>> items = [
                {
                'sys_inventoryItemId':'118312438662692864',
                'meterId': '86IEDD99',
                'dateTime': '2020-01-01T05:50:59Z'
                },
                {
                'sys_inventoryItemId':'118312438662692864',
                'meterId': '45IXZ52',
                'dateTime': '2020-01-07T15:41:14Z'
                }
            ]
        >>> updateItems('meterData', items)
        """
        items = Utils._propertiesToString(items)
        key = f'update{inventoryName}'

        graphQLString= f'''mutation updateItems {{
            {key} (input: 
                {items}
            )
                {{
                    {Utils.errors}           
                    inventoryItems {{
                        sys_inventoryItemId
                }}
            }}
        }}
        '''
        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        if result[key]['errors']:
            Utils._listGraphQlErrors(result, key)

        logger.info(f"{len(result[key]['inventoryItems'])} item(s) updated.")
        return

    def createInventory(
        self, 
        name:str,  
        properties:list, 
        variant:str=None,
        propertyUniqueness:dict=None, 
        historyEnabled:bool=False, 
        hasValitityPeriods:bool=False, 
        isDomainUserType:bool=False
        ) -> str:
        """
        Creates a new inventory. After creation, access rights must be set to add items.
        
        Parameters:
        ----------
        name : str
            Name of the new inventory (only alphanumeric characters allowed, 
            may not begin with a number)
        properties : list
            A list of dicts with the following mandatory keys: 
                name: str
                dataType: enum (STRING, BOOLEAN, DECIMAL, INT, LONG, DATE_TIME, 
                DATE_TIME_OFFSET)
            Optional keys:
                isArray: bool (Default = False)
                nullable: bool (Default = True)
                isReference: bool (Default = False)
                inventoryId: str (mandatory if hasReference = True)
        variant : str
            The inventory variant.
        propertyUniqueness : list
            A list of properties that should be unique in its combination. 
        historyEnabled : bool
            If True, changes in properties will be recorded in item history.
        hasValidityPeriods : bool
            If true, a validity period can be added to the item.    

        Example:
        --------
        >>> propertyDefinitions = [
            {
                'name': 'street',
                'dataType': 'STRING',
                'nullable': False,
            },
            {
                'name': 'postCode',
                'dataType': 'STRING',
                'nullable': False,
            },
            ]
            uniqueness = [{'uniqueKey': 'address', 'properties': ['street', 'postCode']}
        >>> createInventory('appartment', 'propertyDefinitions', propertyUniqueness=uniqueness) 
        """

        _properties = Utils._propertiesToString(properties)

        if variant != None:
            _variantId = f'{Utils._getVariantId(self.variants(), variant)}'
            logger.debug(f"Found variantId: {_variantId}")
            if type(_variantId) != str:
                Utils._error(self, f"Variant name '{name}' not found")
                return

            _variant = f'variantId: "{_variantId}"'
        else: _variant = ''

        if propertyUniqueness != None:
            _propertyUniqueness = 'propertyUniqueness: ' + Utils._uniquenessToString(propertyUniqueness)
            logger.debug(_propertyUniqueness)
        else: _propertyUniqueness = ''

        _history = 'true' if historyEnabled != False else 'false'
        _validityPeriods = 'true' if hasValitityPeriods != False else 'false'
        _domainUser = 'true' if isDomainUserType != False else 'false'
        
        graphQLString= f'''
        mutation createInventory {{
            createInventory (input: {{
                name: "{name}"        
                properties: {_properties}
                {_variant}
                historyEnabled: {_history}
                hasValidityPeriods: {_validityPeriods}
                isDomainUserType: {_domainUser}
                {_propertyUniqueness}
                }})
                {{
                inventory {{
                    inventoryId
                    }}
                {Utils.errors}           
            }}
        }}
        '''
        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        if result['createInventory']['errors']:
            Utils._listGraphQlErrors(result, 'createInventory')
            return

        logger.info(f"New inventory {name} created.")
 
        return result['createInventory']['inventory']['inventoryId']

    def deleteInventory(
        self, 
        inventoryName:str, 
        force:bool=False
        ) -> None:
        """ 
        Deletes an inventory with all its containg items. 
        
        Parameters:
        -----------
        inventoryName : str
            The field name of the inventory.
        force : bool
            Use True to ignore confirmation.

        Example:
        ---------
        >>> deleteInventory('meterData', force=True)
        """

        inventory = self.inventories(where=f'name eq "{inventoryName}"')
        
        if inventory.empty:
            Utils._error(self, f"Unknown inventory '{inventoryName}'.")
            return

        inventoryId = inventory.loc[0, 'inventoryId']
        logger.debug(f'inventoryId: {inventoryId}')

        if force == False:
            confirm = input(f"Press 'y' to delete '{inventoryName}': ")

        key = 'deleteInventory'
        graphQLString= f'''mutation deleteInventory {{
            {key} (input:{{inventoryId: "{inventoryId}"}})
                {{
                    {Utils.errors}           
                }}
            }}
            '''

        if force == True: confirm = 'y'
        if confirm == 'y':
            result = Utils._executeGraphQL(self, graphQLString)
            if result == None: return
            if result[key]['errors']:
                Utils._listGraphQlErrors(result, key)
                return
            else: 
                logger.info(f"Inventory '{inventoryName}' deleted.")
        else: return

    def variants(self) -> pd.DataFrame:
        """
            Returns a dataframe of available variants.
        """

        key = 'variants'
        graphQLString= f'''query getVariants {{
        {key} {{
            name
            variantId
            }}
        }}
        '''
        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        df = pd.json_normalize(result[key])
        return df

    def deleteVariant(
        self, 
        variantId:str, 
        force:bool=False
        ) -> None:
        """Deletes a variant

        Parameters
        -----------
        variantId : str
            The id of the variant.
        force : bool
            Use True to ignore confirmation.
        """


        if force == False:
            confirm = input(f"Press 'y' to delete  variant with Id {variantId}: ")

        key = 'deleteVariant'
        graphQLString= f'''mutation deleteVariant {{
            {key}(input: {{ variantId: "{variantId}" }}) {{
                errors {{
                message
                code
                }}
            }}
        }}
        '''

        if force == True: confirm = 'y'
        if confirm == 'y':
            result = Utils._executeGraphQL(self, graphQLString)
            if result == None: return
        else:
            return

        if result[key]['errors'] != None:
            Utils._listGraphQlErrors(result, key)
        else:
            logger.info(f"Variant deleted.")

    def clearInventory(
        self, 
        inventoryName:str, 
        force:bool=False, 
        pageSize:int=500
        ) -> None:
        """
        Deletes all items from the inventory

        Parameters
        -----------
        inventoryName : str
            The name of the inventory.
        force : bool
            Use True to ignore confirmation.
        pageSize : str = 500
            The number of items to be deleted in a chunk (the maximum number 
            of items that can be deleted in one mutation is restricted).

        Example:
        ---------
        >>> clearInventory('meterData', force=True)
        """
         
        if force == False:
            confirm = input(f"Press 'y' to delete all items in inventory {inventoryName}: ")
        if force == True: confirm = 'y'
        if confirm != 'y':
            return

        count = 0
        lastId = ''

        while True:
            graphQLString= f''' query getItems {{
                    {inventoryName} (
                            pageSize: {pageSize} 
                            {lastId}
                            ) {{
                        sys_inventoryItemId
                    }}
                }}
                '''

            _result = Utils._executeGraphQL(self, graphQLString)
            ids = [item['sys_inventoryItemId'] for item in _result[inventoryName]]
            count += len(ids)
           
            _ids = ''
            for id in ids:
                _ids += f'{{sys_inventoryItemId: "{id}"}}\n'    

            try:
                cursor = _result[inventoryName][-1]['sys_inventoryItemId']
                lastId = f'lastId: "{cursor}"'

                key = f'delete{inventoryName}'
                graphQLString= f'''
                    mutation deleteItems {{
                        {key} ( input: 
                            [{_ids}]
                        )
                        {{
                            {Utils.errors}           
                        }}
                    }}
                '''
                result = Utils._executeGraphQL(self, graphQLString)
            except: 
                break

        if count == 0:
            logger.info(f"Inventory is empty. No items were deleted.")
            return
        if result[key]['errors'] != None:
            Utils._listGraphQlErrors(result, key)
        else:
            logger.info(f"{count} items deleted.")

    def deleteItems(
        self, 
        inventoryName:str, 
        inventoryItemIds:list=None, 
        where:str=None, 
        force=False
        ) -> None:
        """
        Deletes inventory items from a list of inventoryItemIds or by where-criteria. 

        Parameters:
        -----------
        inventoryName: str
            The name of the inventory.
        inventoryItemIds: list = None
            A list of inventoryItemIds that should be deleted.
        where: str = None
            Filter criteria to select items that should be deleted.
        force: bool = False
            Use True to ignore confirmation.

        Examples:
        ---------
        >>> deleteItems('meterData', where='changeDate gt "2020-12-01"', force=True)
        >>> deleteItems('meterData', inventoryItemIds=['ef73d6d5-d1d7-459a-b079-ec640cbb310e'])
        """

        if inventoryItemIds == None and where == None:
            Utils._error(self, f"No list of items and no where-criteria were provided.")
            return
            
            
        if inventoryItemIds != None and where != None:
            logger.warning(f"List of items and where-criteria has been provided. Item list is used.")

        if where != None:
            _result =self.items(inventoryName, fields=['sys_inventoryItemId'], 
                where=where)
            if _result.empty:
                logger.info(f"The where criteria '{where}' led to no results.")
                return

            _inventoryItemIds = list(_result['sys_inventoryItemId'])
            
        if inventoryItemIds != None:
            _inventoryItemIds = inventoryItemIds

        _ids = ''
        for id in _inventoryItemIds:
            _ids += f'{{sys_inventoryItemId: "{id}"}}\n'    
      
        logger.debug(f"GraphQL Ids: {_ids}")
            
        if force == False:
            confirm = input(f"Press 'y' to delete  {len(_inventoryItemIds)} items: ")

        key = f'delete{inventoryName}'
        graphQLString= f'''
            mutation deleteItems {{
                {key} ( input: 
                    [{_ids}]
                )
                {{
                    {Utils.errors}           
                }}
            }}
            '''

        if force == True: confirm = 'y'
        if confirm == 'y':
            result = Utils._executeGraphQL(self, graphQLString)
            if result == None: return
        else:
            return

        if result[key]['errors'] != None:
            Utils._listGraphQlErrors(result, key)
        else:
            logger.info(f"{len(_inventoryItemIds)} items deleted.")

    def updateVariant(
        self, 
        variantName, 
        newName=None, 
        icon=None
        ) -> None:
        """Updates a variant"""

        _variantId = Utils._getVariantId(self.variants(), variantName) 
        logger.debug(f"Found variantId: {_variantId}")

        if newName != None:
            _name = f'name: "{newName}"'
        else:
            _name = 'null'

        if icon != None:
            _icon =f'icon: "{icon}"'
        else:
            _icon = 'null'

        key = 'updateVariant'
        graphQLString = f'''mutation updateVariant {{
            {key} (input:{{
                variantId: "{_variantId}"
                {_name}
                {_icon}
                }})
                {{
                    {Utils.errors}           
                }}
            }}
            '''

        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        return

    def updateArrayProperty(
        self, 
        inventoryName:str, 
        inventoryItemId:str, 
        arrayProperty:str, 
        operation:str, 
        arrayItems:list=None, 
        cascadeDelete:bool=False
        ) -> None:
        """
        Updates a single array property of a single inventoryItemId. Arrays with and without 
        references are supported.

        Parameters:
        -----------
        inventoryName: str
            The name of the inventory where the item is located.
        inventoryItemId: str
            The sys_inventoryItemId of the item.
        arrayProperty: str
            The name of the property whose array items are to be updated.
        operation: str
            The update operation to be performed. Options:
            insert: inserts a list of array items.
            removeById: removes array items by a list of given ids.
            removeByIndex: removes array items by a list of given indices.
            removeAll: removes all array items
        arrayItems: list = None
            A list of indices or item Ids
        cascadeDelete: bool = False
            If array items are refencences, True will delete also the reference items. 

        Examples:
        ---------
        >>> client.updateArrayProperty('meterData', 'A5N45hOOmm',
                'timeSeries', action='insert', arrayItems=['A5FdSjehbE'])
        >>> client.updateArrayProperty('meterData', 'A5N45hOOmm',
                'timeSeries', action='removeByIndex', arrayItems=[0,1], cascadeDelete=True)
        """

        operations = ['insert', 'removeById', 'removeByIndex', 'removeAll']
        if operation not in operations:
            Utils._error(self, f"Action '{operation}' allowed. Possible update operations: {operations}.")
            return

        if operation == 'removeAll':
            try:
                propertyDf = TechStack.inventoryProperties(self, inventoryName)
                arrayProps = propertyDf[propertyDf['name'] == arrayProperty]
                logger.debug(f"Properties of array: {arrayProps}")
                if arrayProps['type'].item() == 'reference':
                    # arrDf = TechStack.items(self, inventoryName, 
                    #     where=f'sys_inventoryItemId eq "{inventoryItemId}"', fields=[f'{arrayProperty}.sys_inventoryItemId'],
                    #     arrayPropertiesToPage=[arrayProperty], arrayPageSize=5000)
                    arrDf = TechStack.items(self, inventoryName, 
                        where=f'sys_inventoryItemId eq "{inventoryItemId}"', fields=[f'{arrayProperty}.sys_inventoryItemId'],
                        )
                else:
                    arrDf = TechStack.items(self, inventoryName, fields=[{arrayProperty}],
                        where=f'sys_inventoryItemId eq "{inventoryItemId}"', arrayPageSize=5000)
            except Exception as err:
                Utils._error(self, err)
                return

            countArray = len(arrDf[arrayProperty].item())
            arrayItems = [num for num in range(countArray)]
            logger.debug(f"Array Items: {arrayItems}")
        _arrayItems = Utils._arrayItemsToString(arrayItems, operation, cascadeDelete)
        logger.debug(f"Array Items as String: {_arrayItems}")

        key = f'update{inventoryName}ArrayProperties'
        graphQLString = f'''mutation updateArray {{
            {key} (
                input: {{
                sys_inventoryItemId: "{inventoryItemId}"
                {arrayProperty}: {{
                    {_arrayItems}
                    }}
                }}
            ) 
            {{
                {Utils.errors}           
            }}
        }}'''

        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        if result[key]['errors'] != None:
            Utils._listGraphQlErrors(result, key)
        else:
            logger.info(f"Array property {arrayProperty} for item {inventoryItemId} updated.")

    def addInventoryProperties(
        self, 
        inventoryName:str, 
        properties:list
        ) -> None:
        """
        Adds one or more inventory properties to an existing inventory.

        Parameters:
        ----------
        inventoryName: str
            Name of inventory
        properties: list
            A list of dicts with the following mandatory keys: 
                name: str
                dataType: enum (STRING, BOOLEAN, DECIMAL, INT, LONG, DATE_TIME, 
                DATE_TIME_OFFSET)
            Optional keys:
                isArray: bool (Default = False)
                nullable: bool (Default = True)
                isReference: bool (Default = False)
                inventoryId: str (mandatory if hasReference = True)
        """

        key = 'addProperty' 
        inventory = self.inventories(where=f'name eq "{inventoryName}"')
        
        if inventory.empty:
            Utils._error(self, f"Unknown inventory '{inventoryName}'.")
            return

        inventoryId = inventory.loc[0, 'inventoryId']
        _properties = Utils._propertiesToString(properties)

        graphQLString= f'''
        mutation {key} {{
        {key} (input: {{
            inventoryId: "{inventoryId}"	
            properties: {_properties}
            }}) 
            {{
                {Utils.errors}           
            }}
        }}
        '''
        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        if result[key]['errors']:
            Utils._listGraphQlErrors(result, key)
            return
        else:
            self.updateClient()
            logger.info(f"New property(ies) added.")

    def updateDisplayValue(
        self, 
        inventoryName:str, 
        displayValue:str
        ) -> None:

        inventoryId = self.structure[inventoryName]['inventoryId']

        key = 'updateInventory'
        graphQLString= f'''mutation updateDisplayValue {{
            {key} (input: {{
                inventoryId: "{inventoryId}"
                displayValue: "{displayValue}"
                }}) {{
                {Utils.errors}
            }}
        }}
        '''
        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        if result[key]['errors']:
            Utils._listGraphQlErrors(result, key)
            return
        else:
            self.updateClient()
            logger.info(f"Display value updated.")
            return

    def updateInventoryName(
        self, 
        inventoryName:str, 
        newName:str
        ) -> None:

        inventoryId = self.structure[inventoryName]['inventoryId']
        key = 'updateInventory'

        graphQLString= f'''mutation updateInventoryName {{
            {key} (input: {{
                name: "{newName}"
                inventoryId: "{inventoryId}"
            }}) {{
                {Utils.errors}
                }}
            }}
        '''
        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return
        if result[key]['errors']:
            Utils._listGraphQlErrors(result, key)
            return
        else:
            logger.info(f"Inventory name updated.")
            self.updateClient()
            return

    def removeProperties(
        self, 
        inventoryName:str, 
        properties:list
        ) -> None:
        """
        Removes a list of properties given as property names. Properties can 
        only be removed if they have no content or are of type 'reference'.

        Parameters:
        -----------
        inventoryName: str
            The name of the inventory where the item is located.
        properties: list
            A list of property names.
        """

        if type(properties) != list:
            properties = [properties]

        propertyList = '['
        for argProperty in properties:
            try:
                propertyId = self.structure[inventoryName]['properties'][argProperty]['propertyId']
                propertyList += f'"{propertyId}",'
            except:
                logger.warning(f"Property '{argProperty}' not found in inventory '{inventoryName}'.")
        propertyList += ']'
            
        inventoryId = self.structure[inventoryName]['inventoryId']
        
        key = 'removeProperty'
        graphQLString= f'''mutation removeProperty {{
            {key} (input: {{
                    inventoryId: "{inventoryId}"
                    propertyIds: {propertyList}
            }}) {{
                {Utils.errors}
                }}
            }}
        '''
        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return
        if result[key]['errors']:
            Utils._listGraphQlErrors(result, key)
            return
        else:
            self.updateClient()
            logger.info(f"Properties have been removed.")
            return

    def updateProperty(
        self, 
        inventoryName:str, 
        propertyName:str, 
        newPropertyName:str=None, 
        nullable:bool=None
        ) -> None:
        
        inventoryId = self.structure[inventoryName]['inventoryId']
        newPropertyName = Utils._argNone('name', newPropertyName)
        nullable = Utils._argNone('nullable', nullable)

        try:
            propertyId = self.structure[inventoryName]['properties'][propertyName]['propertyId']
        except:
            Utils._error(self,f"Property '{propertyName}' not found in inventory '{inventoryName}'.")
            return

        key = 'updateInventory'
        graphQLString= f'''mutation removeProperty {{
            {key} (input: {{
                    inventoryId: "{inventoryId}"
                    properties: [{{
                        propertyId: "{propertyId}"
                        {newPropertyName}
                        {nullable}
                    }}]
            }}) {{
                {Utils.errors}
                }}
            }}
        '''
        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return
        if result[key]['errors']:
            Utils._listGraphQlErrors(result, key)
            return
        else:
            self.updateClient()
            logger.info(f"Property updated.")
            return

    def resync(self) -> None:
        """Resynchronizes read databases"""

        key = 'reSyncReadDatabase'
        graphQLString= f'''mutation resync{{
            {key}
            }}
        '''
        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return
        self.updateClient()

        return
    
    def defaultDataFrame(
        self, 
        maxRows, 
        maxColumns
        ) -> None:
        """Sets default sizes for a DataFrame for the current session"""
        pd.options.display.max_rows = maxRows
        pd.options.display.max_columns = maxColumns
        return
        
    def _convertId(
        self, 
        sys_inventoryItemId:str
        ) -> str:
        """Convers a sys_inventoryItemId into a HAKOM name"""

        graphQLString= f'''mutation convert{{
            convertId(id: "{sys_inventoryItemId}")
            }}
        '''
        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        return result['convertId']

# Some stand alone functions

def encodeBase64(file:str):
    with open(file) as file:
        content = file.read()
        content = base64.b64encode(content.encode('ascii'))
        return content.decode('UTF8')
        
def decodeBase64(content:str):
    return base64.b64decode(content)




from os import name
import pandas as pd
from datetime import datetime, timedelta
from gql import Client
from gql.transport.requests import RequestsHTTPTransport
from loguru import logger

from .utils.ut import Utils
from .utils.ut_autprog import AutProgUtils

class Schedule():
    def __init__(self, accessToken:str, endpoint:str, client:object) -> None:
              
        header = {
            'authorization': 'Bearer ' + Automation.accessToken
        }
        
        transport =  RequestsHTTPTransport(url=endpoint, headers=header, verify=True, proxies=client.proxies)
        self.client = Client(transport=transport, fetch_schema_from_transport=False)
        self.raiseException = client.raiseException
        self.defaults = client.defaults

    def schedules(
        self, 
        fields:list=None, 
        where:str=None) -> pd.DataFrame:
        """
        Returns schedules in a DataDrame

        Parameters:
        -----------
        fields: list | str = None
            A list of all properties to be queried. If None, all properties will be queried.
        where: str = None
            Use a string to add where criteria like
            ''workflowId eq "meteoData"'.

        Example:
        --------
        >>> Schedule.schedules(
                where='workflowId == "meteoData"', 
                fields=['name', 'cron']
            )   
        """

        key = 'schedules'

        if fields != None:
            if type(fields) != list:
                fields = [fields]
            _fields = Utils._queryFields(fields, recursive=True)   
        else:
            _fields =f'''scheduleId
                name
                description
                workflowId
                businessKey
                cron
                isActive
                nextFireTime
                variables {{
                    key
                    value
                }}''' 

        resolvedFilter = ''
        if where != None: 
            resolvedFilter = f'({Utils._resolveWhereString(where)})'

        graphQLString = f'''query schedules {{
            {key}{resolvedFilter}  {{
                {_fields}
            }}
        }}
        '''

        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        df = pd.json_normalize(result[key])
        return df

    def createSchedule(self, name:str, workflowId:str, businessKey:str, cron:str, 
        isActive:bool=True, description:str=None, variables:dict=None) -> str:
        """Creates a schedule and returns the schedule Id"""

        if isActive == True:
            isActive = 'true'
        else:
            isActive = 'false'

        if description != None:
            description = description
        else: 
            description = ''

        if variables != None:
            _variables = 'variables: [\n'
            for k,v in variables.items():
                _variables += f'{{key: "{k}", value: "{v}"}}\n'
            _variables += ']'
        else: _variables = ''

        graphQLString = f'''mutation createSchedule {{
            createSchedule(input:{{
                name: "{name}"
                workflowId: "{workflowId}"
                businessKey: "{businessKey}"
                cron: "{cron}"
                description: "{description}"
                isActive: {isActive}
                {_variables}              
            }})
            {{
                schedule {{
                    scheduleId
                }}
                errors {{
                    message
                }}
            }}
        }}'''

        result = Utils._executeGraphQL(self, graphQLString)
        logger.debug(graphQLString)
        if result == None: return

        key = 'createSchedule'
        if result[key]['errors']:
            Utils._listGraphQlErrors(result, key)
        else:
            scheduleId = result[key]['schedule']['scheduleId']
            logger.info(f"New schedule {scheduleId} created.")

        return scheduleId

    def updateSchedule(self, scheduleId, name:str=None, workflowId:str=None, businessKey:str=None, 
        cron:str=None, isActive:bool=None, description:str=None, variables:dict=None) -> None:
        """
        Updates a schedule. Only arguments that ar not None will overwrite respective fields.
        
        Parameters:
        -----------
        scheduleId : str
            The Id of the schedule that is to be updated.
        name : str
            The name of the schedule.
        workflowId : str
            The Id of the workflow that shall be executed with this schedule.
        cron : str
            The cron expression. For detailed information loop up
            http://www.quartz-scheduler.org/documentation/quartz-2.3.0/tutorials/crontrigger.html
        isActive : bool
            Determines, if the schedule should execute the workflow or not.
        description : str
            A description of the schedule.
        variables : dict
            A dictionary of variables that are used by tasks in the workflow.

        Example:
        --------
        >>> vars = {
                'var1': 99,
                'var2': "AnyString"
            }
        >>> client.Scheduler.updateSchedule('112880211090997248', name='test_schedule',
                isActive=True, variables=vars)
        
        """

        updateScheduleArgs = ''

        if name != None: updateScheduleArgs += f'name: "{name}"\n'
        if workflowId != None: updateScheduleArgs += f'workflowId: "{workflowId}"\n'
        if businessKey != None: updateScheduleArgs += f'businessKey: "{businessKey}"\n'
        if cron != None: updateScheduleArgs += f'cron: "{cron}"\n'
        if isActive != None: updateScheduleArgs += f'isActive: {str(isActive).lower()}\n'
        if description != None: updateScheduleArgs +=  f'description: "{description}"\n'

        if variables != None:
            _variables = 'variables: [\n'
            for k,v in variables.items():
                _variables += f'{{key: "{k}", value: "{v}"}}\n'
            _variables += ']'
            updateScheduleArgs += _variables

        graphQLString = f'''mutation updateSchedule {{
            updateSchedule(
                scheduleId: "{scheduleId}"
                input:{{
                    {updateScheduleArgs}
            }})
            {{
                errors {{
                    message
                }}
            }}
        }}'''

        result = Utils._executeGraphQL(self, graphQLString)
        logger.debug(graphQLString)
        if result == None: return

        key = 'updateSchedule'
        if result[key]['errors']:
            Utils._listGraphQlErrors(result, key)
        else:
            logger.info(f"Schedule {scheduleId} updated.")

        return

    def deleteSchedule(self, scheduleId:str, force:bool=False):
        """Deletes a schedule"""

        if force == False:
            confirm = input(f"Press 'y' to delete schedule '{scheduleId}': ")

        graphQLString = f'''mutation deleteSchedule {{
            deleteSchedule (scheduleId: "{scheduleId}")
            {{
                errors {{
                message
                }}
            }}
        }}
        '''      

        if force == True: confirm = 'y'
        if confirm == 'y':
            result = Utils._executeGraphQL(self, graphQLString)
            if result == None: return

            key = 'deleteSchedule'
            if result[key]['errors']:
                Utils._listGraphQlErrors(result, key)
            else:
                logger.info(f"Schedule {scheduleId} deleted")
                return None

    def nextFireTimes(self, workflowId:str, fromTimepoint:str=None, toTimepoint:str=None, count:int=None):
        """Show next fire times of a workflow"""

        if fromTimepoint == None:
            fromTimepoint = datetime.today().isoformat()

        if toTimepoint == None:
            toTimepoint = datetime.today() + timedelta(days=3)

        if count == None:
            _count = ''
        else: 
            _count = f'count: {count}'

        graphQLString = f'''query nextFireTimes {{
            nextFireTimes (
                workflowId: "{workflowId}",
                from: "{fromTimepoint}", 
                to: "{toTimepoint}",
                {_count}
                
                ) {{
                scheduleId
                fireTime
                }}
            }}
        '''
        
        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        df = pd.json_normalize(result['nextFireTimes'])

        return df

class Automation():
    accessToken = ''

    def __init__(self, accessToken:str, endpoint:str, client:object) -> None:
        
        Automation.accessToken = accessToken
        
        header = {
            'authorization': 'Bearer ' + accessToken
        }
        
        transport =  RequestsHTTPTransport(url=endpoint, headers=header, verify=False, proxies=client.proxies)
        self.client = Client(transport=transport, fetch_schema_from_transport=True)
        self.raiseException = client.raiseException
        self.defaults = client.defaults

    def workflows(self) -> pd.DataFrame:
        """Returns a DataFrame of all Workflows"""

        graphQLString = f'''query workflows {{
            workflows {{
                id
                name
                description
                }}
            }}
            '''

        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        df = pd.json_normalize(result['workflows'])
        return df

    def workflowInstances(self, workflowId:str=None, fromTimepoint:str=None, toTimepoint:str=None, 
        fields:list=None, where:str=None, showTasks=False) -> pd.DataFrame:
        """Shows Instances of a workflow. If workflowId=None, all Instances of all 
        workflows will be returned."""

        meta = ['id', 'name', 'businessKey', 'version', 'startTime', 'endTime', 'state']
        key = 'workflowInstances'

        if workflowId != None:
            _workflowId = f'workflowId: "{workflowId}"'
        else:
            _workflowId = ''

        if fromTimepoint != None:
            _fromTimepoint = f'from: "{fromTimepoint}"'
        else: 
            _fromTimepoint = ''

        if toTimepoint != None:
            _toTimepoint = f'from: "{toTimepoint}"'
        else: 
            _toTimepoint = ''

        if fields != None:
            if type(fields) != list:
                fields = [fields]
            _fields = Utils._queryFields(fields, recursive=True)   
        else:
            _fields =f'''
                id
                name
                businessKey
                version
                startTime
                endTime
                state
                variables {{
                    name
                    value
                    time
                }}''' 

        resolvedFilter = ''
        if where != None: 
            resolvedFilter = Utils._resolveWhereString(where)

        if showTasks != False:
            _tasks = f'''tasks {{
                            id
                            topic
                            workerId
                            timestamp
                            state
                            retries
                            errorMessage
                        }}'''
        else:
            _tasks = ''

        graphQLString = f'''query Instances {{
            {key}({_workflowId}, {_fromTimepoint}, {_toTimepoint}, {resolvedFilter}, all:true) {{
                {_fields}
                {_tasks}
                }}
            }}
            '''
     
        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        if showTasks != False:
            df = pd.json_normalize(result[key], meta=meta, record_path=['tasks'], record_prefix='task.', errors='ignore')
            if 'startTime' in df.columns:
                df = df.sort_values(by='startTime', ascending=False)
        else:
            df = pd.json_normalize(result[key])
            if 'startTime' in df.columns:
                df = df.sort_values(by='startTime', ascending=False)
        return df
    
    def createWorkflow(self, id, name, description:str=None):

        graphQLString = f'''mutation createWorkflow {{
            createWorkflow(
                input: {{
                    id: "{id}"
                    name: "{name}"
                    description: "{description}"
                }}
                ) {{
                    ...on CreateWorkflowError {{
                    message
                    }}
                    ... on WorkflowCreated {{
                        workflow {{
                            id
                        }}
                    }}
                }}
            }}
        '''
        
        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        return result
                        
    def deployWorkflow(self, workflowId:str, filePath:str) -> None:
        """Deploys a Camunda XML to an existing workflow"""

        fileContent = Utils._encodeBase64(filePath)
        logger.debug(f"fileContent: {fileContent[:10]}")

        graphQLString = f'''mutation deployWorkflow {{
            deployWorkflow(
                input: {{
                    fileContentBase64: "{fileContent}"
                    workflowId: "{workflowId}"
                }}
            ) {{
                ... on DeployWorkflowError {{
                    message
                }}
                ... on InvalidWorkflowProcessId {{
                    processId
                    workflowId
                    message
                }}
                ... on WorkflowDeployed {{
                    version
                }}
            }}
        }}
        '''

        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        return result

    def startWorkflow(self, workflowId:str, businessKey:str, inputVariables:dict=None):
        """Starts a workflow"""

        if inputVariables == None:
            _vars = ''
        else:
            _vars = AutProgUtils._varsToString(inputVariables, 'input')

        graphQLString = f'''
            mutation ExecuteWF {{
                startWorkflow(input: {{ 
                    businessKey: "{businessKey}"
                    workflowId: "{workflowId}" 
                    {_vars}
                    }}
                ) {{
                    ... on ProcessDefinitionNotFound {{
                        workflowId
                        message
                        }}
                    ... on StartWorkflowError {{
                            message
                            }}
                    ... on WorkflowStarted {{
                        workflowInstanceId
                        }}
                    }}
                }}
            '''

        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        return result

    def deleteWorkflow(self, workflowId:str):
        """Deletes a workflow"""

        graphQLString = f'''mutation deleteWorkflow {{
            deleteWorkflow (id: "{workflowId}")
            {{
                ... on DeleteWorkflowError {{
                    message
                    }}
                ...on WorkflowDeleted {{
                    success
                    }}
                ... on WorkflowNotFound {{
                    workflowId
                    message
                    }}
                
                }}
            }}
        '''

        result = Utils._executeGraphQL(self, graphQLString)
        return result

    def terminateWorkflowInstance(self, workflowInstanceId):
        """Terminates a workflow instance"""

        graphQLString = f'''mutation terminateWorkflowInstance {{
            terminateWorkflowInstance(
                workflowInstanceId:"{workflowInstanceId}") {{
                ...on TerminateWorkflowInstanceError {{
                    message
                    }}
                ...on WorkflowInstanceNotFound {{
                    workflowInstanceId
                    message
                    }}
                ...on WorkflowInstanceTerminated {{
                    success
                    }}
                }}
            }}
        '''

        result = Utils._executeGraphQL(self, graphQLString)
        return result

    def updateWorkflow(self, workflowId:str, name:str=None, description:str=None):
        """Updates a workflow (name and description can be changed)"""

        name = Utils._argNone('name', name)
        description = Utils._argNone('description', description)

        key = 'updateWorkflow'
        graphQLString = f'''mutation updateWorkflow {{
            {key}(workflowId: "{workflowId}", properties: {{
                {description}
                {name}
                }}) {{
                    ... on UpdateWorkflowError {{
                    message
                    }}
                    ... on WorkflowNotFound {{
                    workflowId
                    message
                    }}
                    ... on WorkflowUpdated {{
                    workflow {{
                        id
                        name
                        description
                        }}
                    }}
                }}
            }}
            '''

        result = Utils._executeGraphQL(self, graphQLString)
        return result

    def retryTask(self, externalTaskId):
        key = 'retryTask'

        graphQLString = f'''mutation retryTask {{
            {key}(externalTaskId: "{externalTaskId}") {{
                ... on RetryTaskError {{
                message
                }}
                ...on TaskNotFound {{
                message
                }}
                ... on TaskRetried {{
                success
                }}
            }}
        }}
        '''

        result = Utils._executeGraphQL(self, graphQLString)
        return result



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / 'README.md').read_text(encoding='utf-8')

setup(name='seven2one',
      version='2.34.0',
      description='Functions to interact with the Seven2one TechStack',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='http://www.seven2one.de',
      author='Seven2one Informationssysteme GmbH',
      author_email='info@seven2one.de',
      license='MIT',
      packages=['seven2one', 'seven2one.utils'],
      include_package_data=True,
      install_requires=[
            'pandas', 'gql==3.0.0', 'pytz', 'tzlocal', 'pyperclip', 'loguru', 'requests', 'requests_toolbelt'
            ],
      classifiers =[
            'Development Status :: 3 - Alpha',
            'Natural Language :: English',
            'Operating System :: OS Independent',
            ],
      python_requires='>=3.6',
      zip_safe=False)



from dataclasses import dataclass
from typing import Union
import pytz
import os
import base64
import math
import pandas as pd
from gql import gql
from tzlocal import get_localzone
from loguru import logger
from collections import namedtuple

if os.name == 'nt': #only if the platform is Windows, pyperclip will be imported
    import pyperclip

@dataclass
class Defaults():
    timeZone: str = 'UTC'
    useDateTimeOffset: bool = True
    copyGraphQLString: bool = False
    includeSystemProperties: bool = False

class Structure():
    queryStructure = f'''query inventories {{
    inventories (pageSize:1000) {{
        name
        inventoryId
        displayValue
        isDomainUserType
        hasValidityPeriods
        historyEnabled
        propertyUniqueness {{
            uniqueKey
            properties
            }}
        variant {{
            name
            properties {{
                name
                type
                isArray
                nullable
                }}
            }}
        properties {{
            name
            ...Scalar
            type
            isArray
            nullable
            propertyId
            ... Reference 
            }}
        }}
    }}
    fragment Scalar on IScalarProperty {{
        dataType
        }}
    fragment Reference on IReferenceProperty {{
        inventoryId
        inventoryName
        }}
    '''

    def _introspectionQueryString():
        introspectionQueryString = r'''
            query IntrospectionQuery { __schema { queryType { name } mutationType 
                { name } subscriptionType { name } types { ...FullType } directives
                { name description locations args { ...InputValue } } } }

            fragment FullType on __Type { kind name description fields(includeDeprecated: true) { name description args 
                { ...InputValue } type { ...TypeRef } isDeprecated deprecationReason } inputFields { ...InputValue } interfaces 
                { ...TypeRef } enumValues(includeDeprecated: true) { name  } possibleTypes { ...TypeRef } } 

            fragment InputValue on __InputValue { name description type { ...TypeRef } defaultValue } 
            fragment TypeRef on __Type { kind name ofType { kind name ofType { kind name ofType 
                    { kind name ofType { kind name ofType { kind name ofType { kind name ofType { kind name } } } } } } } }
        '''
        return introspectionQueryString

    def _fullStructureDict(structure) -> dict:
        """Converts the query of all inventories with all fields into a pure dict"""

        def subdict(inputObject, name):
            itemDict = {}
            for item in inputObject:
                itemDict.setdefault(item[name], {})
                for k, v in item.items():
                    itemDict[item[name]].setdefault(k,v)
            return itemDict

        structureDict = {}
        for inventory in structure['inventories']:
            inventoryName = inventory['name']
            structureDict.setdefault(inventoryName, {})
            for definitionKey, definitionValue in inventory.items():
                if not isinstance(definitionValue, (list, dict)):
                    structureDict[inventoryName].setdefault(definitionKey, definitionValue)
                else:
                    if definitionKey == 'properties':
                        subDict = subdict(inventory[definitionKey], 'name')
                        structureDict[inventoryName].setdefault(definitionKey, subDict)
                    if definitionKey == 'propertyUniqueness':
                        subDict = subdict(inventory[definitionKey], 'uniqueKey')
                        structureDict[inventoryName].setdefault(definitionKey, subDict)
                    if definitionKey == 'variant':
                        structureDict[inventoryName].setdefault(definitionKey, {})
                        structureDict[inventoryName][definitionKey].setdefault('name', definitionValue['name'])
                        subDict = subdict(inventory[definitionKey]['properties'], 'name')
                        structureDict[inventoryName][definitionKey].setdefault('properties', subDict)
        return structureDict

    def _fullStructureNT(structure:dict) -> namedtuple:
        """
        Provides the complete data structure of dynamic objects as named tuple. 
        Needs structureDict first
        """
        def _subItem(object:dict):
            Item = namedtuple('Item', object.keys())
            itemDict = {}
            for key, value in object.items():
                if isinstance(value, dict):
                    subItem = _subItem(value)
                    itemDict.setdefault(key, subItem)
                else:
                    itemDict.setdefault(key, value)
            item = Item(**itemDict)
            return item

        Item = namedtuple('Item', structure.keys())
        itemDict = {}
        for key, value in structure.items():
            if isinstance(value, dict):
                subItem = _subItem(value)
                itemDict.setdefault(key, subItem)
            else:
                itemDict.setdefault(key, value)
        return Item(**itemDict)

    def _inventoryNT(structure) -> namedtuple:
        """
        Provides a simplified namedtuple of dynamic objects for interactive usage
        """
        inventoryDict = {key:key for key in structure.keys()}
        Inventories = namedtuple('Inventories', inventoryDict.keys())
        return Inventories(**inventoryDict)

    def _inventoryPropertyNT(structure) -> namedtuple:
        """
        Provides a simplified namedtuple of inventory properties for interactive usage
        """
        Inventory = namedtuple('Inventories', structure.keys())
        inventoryDict = {}

        for inventory in structure.keys():
            propertyDict = {}
            for key in structure[inventory]['properties'].keys():
                propertyDict.setdefault(key, key)
                Properties = namedtuple('Properties', propertyDict.keys())
                properties = Properties(**propertyDict)
            inventoryDict.setdefault(inventory, properties)
        return Inventory(**inventoryDict)

class Utils():
    errors =  f'''errors {{
                    message
                    code
                }}'''

    def _error(self, msg:str):
        if self.raiseException: raise Exception(msg)
        else:
            logger.error(msg)
            return

    def _timeZone(timeZone):
        if timeZone=='local': 
            localTimeZone = get_localzone().zone
            return str(pytz.timezone(localTimeZone))
        else:
            return str(pytz.timezone(timeZone))
    
    def _graphQLList(itemList:list) -> str:

        def is_number(n):
            try:
                float(n)
            except ValueError:
                return False
            return True

        """Converts a list to a graphQL list"""
        result = '['
        for item in itemList:
            if is_number(item):
                result += f'{item},'
            else:
                result += f'"{item}",'
        result += ']'
        return result

    def _queryFields(fieldList:list, arrayTypeFields:list=None, arrayPageSize:int=None, filter:dict=None, recursive=False) -> str:
        """
        Transforms a Python list of fields into graphQL String of fields
        including fields of referenced inventories        
        """

        def nestedItem(item):
            nonlocal fields
            line = ''
            for i in range(itemLength - 1):
                if arrayTypeFields != None and item[i] in arrayTypeFields:
                    if filter != None and item[i] in filter.keys():
                        line += f'{item[i]} (pageSize: {arrayPageSize}, where: {filter[item[i]]}) {{ '
                    else:
                        line += f'{item[i]} (pageSize: {arrayPageSize}) {{ '
                else:
                    line += f'{item[i]} {{ '
            line += f'{item[-1]} '
            for _ in range(itemLength - 1):
                line += '}'
            line += ' \n'
            fields += line

        fields = ''
        splitList = [item.split('.') for item in fieldList]
        logger.debug(f"intermediate step - splitted list: {splitList}")
    
        for item in splitList:
            if len(item) == 1:
                fields += f'{item[0]}  \n'
            else:
                itemLength = len(item)
                if recursive == False:
                    if item[-1] == '_displayValue':  
                        nestedItem(item)
                    if item[-1] == 'sys_inventoryItemId':  
                        nestedItem(item)
                else:
                    nestedItem(item)
        return fields

    def _handleWhere(self, where, inventoryName:str=None) -> str:
        """Handles the where argument if any"""
        if where != None: 
            resolvedFilter = Utils._resolveWhere(self, where, inventoryName)
            logger.debug(f"Resolved filter: {resolvedFilter}")
            if 'topLevel' in resolvedFilter.keys():
                topLevelFilter = resolvedFilter['topLevel']
                if topLevelFilter == None: return
        else:
            topLevelFilter = ''
            resolvedFilter = None
        if '[]' in topLevelFilter:
            Utils._error(self, "Filter is not resolvable.")
            return ''
        return topLevelFilter, resolvedFilter

    def _toGraphQL(value) -> Union[str, int, float]:
        """
        Escapes a value for usage in GraphQL. Surrounding " will be added automatically.

        Parameters
            value
                The value which should be escaped for GraphQL.
        
        Returns
            Escaped value. Includes surrounding " for strings.

        Raises
            ValueError
                If value is NaN.

        Examples:
        * string ABC becomes string "ABC"
        * string A"BC becomes string "A\\"BC"
        * string "ABC" becomes string "\\"ABC\\""
        * int 123 becomes 123
        * float 12.34 becomes 12.34
        * None becomes string null
        * True becomes string true
        * False becomes string false
        * [1, 2, 3] becomes [1, 2, 3]
        """
        if (value == None):
            return 'null'
        if (isinstance(value, bool)):
            return 'true' if value else 'false'
        if (isinstance(value, int)):
            return str(value);
        if (isinstance(value, float)):
            if (math.isnan(value)):
                raise ValueError('Value is NaN.')
            return str(value)
        if type(value) == list:
            escapedList = ', '.join(Utils._toGraphQL(item) for item in value)
            return f'[{escapedList}]'
        escapedValue = str(value).replace('\\', '\\\\').replace('"', '\\"')
        return f'"{escapedValue}"'

    def _resolveWhere(self, filterArg, inventoryName=None):
        """
        How does this work:
        A list of lists is created, where 'or terms' are the elements of the parent list and 
        'and terms' are elements of a child list of an or (or single) element.
        For each 'and' in a child list, the string will be closed with an extra '}'

        Lists (as string) are treated seperately, but work the same as a single or an 'and' element.
        """

        def mapOperator(operator):
            operators = {
                '==': 'eq',
                'eq': 'eq',
                'in': 'in',
                '<': 'lt',
                '>': 'gt',
                '<=': 'lte',
                '>=': 'gte',
                'lt': 'lt',
                'gt': 'gt',
                'lte': 'lte',
                'gte': 'gte',
                'contains': 'contains',
                '!=': 'neq',
                'ne': 'neq',
                'neq': 'neq',
                'not in': 'not in',
                'startswith': 'startswith',
                'endswith': 'endswith',
                '=': 'eq'
            }
            if operator in operators:
                return operators[operator]
            else:
                logger.error(f"Unknown operator '{operator}'")

        def _split(itemToSplit:str) -> list:
            "Splits a string by spaces and strips all blanks"
            splitted = itemToSplit.split(' ')
            for item in reversed(splitted): 
                if item == '': 
                    splitted.remove(item)

            if len(splitted) > 3:
                Utils._error(self, f'''Invalid filter criteria {itemToSplit}. Did you miss to put the search string in double quotes ("")?''')
            return splitted

        def _secondLevelFilter(el:namedtuple, inventoryName:str, resultDict:dict):
            """Builds the filter part for second level"""

            subProperty = el.property.split('.')[0]
            subFilterProperty = el.property.split('.')[1]
            try:
                subInventory = self.structure[inventoryName]['properties'][subProperty]['inventoryName']
            except:
                Utils._error(self, f"Unknown property '{subProperty}' in '{el.property}'.")
                return

            df = self.items(subInventory, fields=['sys_inventoryItemId'], 
                where=[f'{subFilterProperty} {el.operator} {el.searchItem}'])

            try:
                itemIds = list(df['sys_inventoryItemId'])
            except:
                logger.warning(f"'{subFilterProperty} {el.operator} {el.searchItem}' does not lead to a result.")
                return ''
            if len(itemIds) == 0:
                logger.error(f"{el.property} is not a valid filter criteria.")
                return ''

            # Check, if filter compund is of array type (treated differently)
            subFilterPropertyIsArray = self.structure[inventoryName]['properties'][subProperty]['isArray']
            if subFilterPropertyIsArray:
                if len(itemIds) == 1:
                    whereString = f'{{sys_inventoryItemId: {{ {mapOperator(el.operator)}: "{itemIds[0]}" }} }}'
                else: # if it is a list
                    itemIds = Utils._graphQLList(itemIds)
                    whereString = f'{{sys_inventoryItemId: {{ in: {itemIds} }} }}'
            else:
                if len(itemIds) == 1:
                    whereString = f'{{{subProperty}: {{sys_inventoryItemId: {{ {mapOperator(el.operator)}: "{itemIds[0]}" }} }} }}'
                else: # if it is a list
                    itemIds = Utils._graphQLList(itemIds)
                    whereString = f'{{{subProperty}: {{sys_inventoryItemId: {{ in: {itemIds} }} }} }}'
            
            #subFilterPropertyIsArray = self.structure[inventoryName]['properties'][subProperty]['isArray']
            if subFilterPropertyIsArray:
                resultDict.setdefault(subProperty, whereString)
                logger.warning("Test")
                Utils._error(self, f"Array properties are not yet supported in filters: '{subProperty}'")
                return resultDict.setdefault(subProperty, whereString)
                
            else:
                return whereString    

        def createFilterPart(subElement) -> namedtuple:
            SubElement = namedtuple('SubElement', ['property', 'operator', 'searchItem'])
            if '[' in subElement: #if it is a list
                x = subElement.find('[')
                sub = subElement[:x].split(' ')
                sub_last = subElement[x:]
                if sub_last.count('"') == 0: #if it is a real list put in by {} 
                    sub_last = sub_last[1:-1].split(',')
                    sub_last = [item.lstrip() for item in sub_last]
                    sub_last = [item.replace("'","") for item in sub_last]
                    sub_last = Utils._graphQLList(sub_last)
                el = SubElement(property=sub[0], operator=sub[1], searchItem=sub_last)
                if el.property.count('.') == 1:
                    whereString = _secondLevelFilter(el, inventoryName, resultDict)
                else: 
                    whereString = f'{{ {el.property}: {{ {mapOperator(el.operator)}: {el.searchItem} }} }}'
                return whereString
            else:
                if subElement.count('"') == 2: # search item is a string
                    x = subElement.find('"')
                    sItem = subElement[x:]
                    sub = _split(subElement[:x])
                elif subElement.count('"') == 0: # search item is not a string (but valid)
                    sub = _split(subElement)
                    sItem = sub[2]
                else: # invalid sub elements with 1 or more than 3 double quotes
                    Utils._error(self, f"Invalid filter criteria {subElement}")
        
                el = SubElement(property=sub[0], operator=sub[1], searchItem=sItem)
                if el.property.count('.') == 1:
                    whereString = _secondLevelFilter(el, inventoryName, resultDict)
                else: 
                    whereString = f'{{ {el.property}: {{ {mapOperator(el.operator)}: {el.searchItem} }} }}'
                return whereString

        def filterPart(arg):
            if type(arg) == str:
                return createFilterPart(arg)
            elif type(arg) == list:
                whereString = '{and: ['
                for i, subArg in enumerate(arg):
                    whereString += filterPart(subArg)
                    if i != len(arg) - 1:
                        whereString += ', '
                whereString += ']} '
                return whereString
            elif type(arg) == tuple:
                whereString = '{or: ['
                for i, subArg in enumerate(arg):
                    whereString += filterPart(subArg)
                    if i != len(arg) - 1:
                        whereString += ', '
                whereString += ']} '
                return whereString
            elif type(arg) == bool:
                Utils._error(self, f"Wrong syntax of filter '{filterArg}'.")
                return

        resultDict = {}
        

        if type(filterArg) in [list, tuple]:
            whereString = filterPart(filterArg)
        else:
            if ' or ' in filterArg:
                Utils._error(self, f"'or' is not supported in string type filters.")             
            filterArg = filterArg.split(' and ')
            whereString = filterPart(filterArg)
            resultDict.setdefault('topLevel', 'where: ' + whereString)

        if whereString:
            resultDict.setdefault('topLevel', 'where: ' + whereString)
            
        logger.debug(f"ResultDict: {resultDict}")
        return resultDict


    def _propertiesToString(properties:list) -> str:
        """ Converts a list of property dicts for many items into a string """
        if type(properties) == list:
            _properties = '[\n'
            for property in properties:
                _properties += '{\n'
                for key, value in property.items():
                    _properties += Utils._customProperties(key, value)      
                
                _properties += '}\n'
            _properties += ']'
            return _properties
        if type(properties) == dict:
            _properties = '{\n'
            for key, value in properties.items():
                _properties += Utils._customProperties(key, value)
            _properties += '}\n'
            return _properties
        else:
            logger.error(f"Type of property items has to be either list ord dict.")
            return

    def _tsPropertiesToString(properties:list) -> str:
        """ Converts a list of property dicts for many items into a string """
        timeUnit, factor = 'timeUnit', 'factor'
        _properties = '[\n'
        for property in properties:
            _properties += '{\n'
            for key, value in property.items():
                if key == 'resolution':
                    try:
                        _properties += f'{key}: {{\n'
                        _properties += f'timeUnit: {value[timeUnit]}\n'
                        _properties += f'factor: {value[factor]}\n'
                        _properties += f'}}\n'
                    except KeyError:
                        logger.error("Missing 'timeUnit' and/or 'factor' for Timeseries resolution")
                        return
                else:
                    _properties += Utils._customProperties(key, value)
            
            _properties += '}\n'
        _properties += ']'
        return _properties

    def _addToGroupPropertiesToString(groupItemId:str, properties:list) -> str:
        """ Converts a list of property dicts for many items into a string """
        
        _properties = '[\n'
        for property in properties:
            _properties += f'{{sys_groupInventoryItemId: "{groupItemId}"\n'
            for key, value in property.items():
                _properties += Utils._customProperties(key, value)      
            
            _properties += '}\n'
        _properties += ']'
        return _properties

    def _uniquenessToString(propertyUniqueness: list):
        """
        Converts a list of unique keys into a string
        """

        _uniqueKeys = '[\n'
        for item in propertyUniqueness:
            key = item['uniqueKey']
            _uniqueKeys += f'{{uniqueKey: "{key}" properties: ['
            for value in item['properties']:
                _uniqueKeys += f'"{value}",'     
            
            _uniqueKeys += ']}\n'
        _uniqueKeys += ']'
        return _uniqueKeys

    def _customProperties(key: str, value: object) -> str:
        """
        Used internally (in Utils) as helper function
        """

        if key == 'dataType':
            return f'{key}: {value}\n'
        return  f'{key}: {Utils._toGraphQL(value)}\n'

    def _properties(scheme, inventoryName:str, recursive:bool=True, sysProperties:bool=False) -> dict:
        """
        Creates a nested (or unnested) dict with properties and arra 
        type fields for further usage out of the scheme
        """

        propertyDict = {'properties':{}, 'arrayTypeFields':[]}
        properties = propertyDict['properties']
        arrayTypeFields = propertyDict['arrayTypeFields']

        def _getInventoryObject(scheme, inventoryName):
            for item in scheme['__schema']['types']:
                if item['name'] == inventoryName:
                    return item['fields']
        
        def _createSubDict(subInv):
            subDict = {}
            inventoryObject = _getInventoryObject(scheme, subInv)
            for item in inventoryObject:
                if sysProperties == False:
                    if item['name'].startswith('sys_'):
                        if item['name'] == 'sys_inventoryItemId': pass
                        else: continue
                if item['type']['kind'] == 'SCALAR':
                    subDict.setdefault(item['name'], item['type']['name'])
                if item['type']['kind'] == 'LIST':
                    if item['name'] == 'sys_permissions': pass
                    elif item['type']['ofType']['kind'] == 'OBJECT':
                        arrayTypeFields.append(item['name'])
                        if recursive == False:
                            subDict.setdefault(item['name'], item['type']['name'])
                        else:
                            subDict.setdefault(item['name'], _createSubDict(item['type']['ofType']['name']))
                    else:
                        arrayTypeFields.append(item['name'])
                        subDict.setdefault(item['name'], item['type']['name'])
                if item['type']['kind'] == 'OBJECT':
                    if recursive == False:
                        subDict.setdefault(item['name'], item['type']['name'])
                    else:
                        subDict.setdefault(item['name'], _createSubDict(item['type']['name']))

            return subDict

        inventoryObject = _getInventoryObject(scheme, inventoryName)
        for item in inventoryObject:
            if sysProperties == False:
                if item['name'].startswith('sys_'):
                    if item['name'] == 'sys_inventoryItemId': pass
                    else: continue
            if item['type']['kind'] == 'SCALAR':
                properties.setdefault(item['name'], item['type']['name'])
            if item['type']['kind'] == 'LIST':
                if item['name'] == 'sys_permissions': pass
                elif item['type']['ofType']['kind'] == 'OBJECT':
                    arrayTypeFields.append(item['name'])
                    if recursive == False:
                        properties.setdefault(item['name'], item['type']['name'])
                    else:
                        properties.setdefault(item['name'], _createSubDict(item['type']['ofType']['name']))
                else:
                    arrayTypeFields.append(item['name'])
                    properties.setdefault(item['name'], item['type']['name'])
            if item['type']['kind'] == 'OBJECT':
                if recursive == False:
                    properties.setdefault(item['name'], item['type']['name'])
                else:
                    properties.setdefault(item['name'], _createSubDict(item['type']['name']))

        logger.debug(f"returned property dict: {propertyDict}")
        
        return propertyDict


    def _propertyList(propertyDict:dict, recursive:bool=False) -> list:
        """Uses _properties() to create a flat list of properties"""

        propertyList = []
    
        def instDict(subDict, path):
            if recursive == True:
                for k, v in subDict.items():
                    if isinstance(v, dict):
                        path = f'{path}.{k}'
                        instDict(subDict[k], path)
                    else:
                        propertyList.append(f'{path}.{k}')
            else:
                if '_displayValue' in subDict.keys(): #if it is not an array
                    propertyList.append(f'{path}._displayValue')
                else: # if it is an array
                    propertyList.append(f'{path}.sys_inventoryItemId')

        for k, v in propertyDict.items():
            if isinstance(v, dict):
                instDict(propertyDict[k], k)
            else:
                propertyList.append(k)

        logger.debug(f'returend property list: {propertyList}')
        return propertyList
            
    def _propertyTypes(propertyDict:dict) -> dict:
        """Uses _properties() o create a flat dictionary of properties"""

        propertyTypes = {}
        
        def instDict(subDict, path):
            for k, v in subDict.items():
                if isinstance(v, dict):
                    path = f'{path}.{k}'
                    instDict(subDict[k], path)
                else:
                    propertyTypes.setdefault(f'{path}.{k}', v)

        for k, v in propertyDict.items():
            if isinstance(v, dict):
                instDict(propertyDict[k], k)
            else:
                propertyTypes.setdefault(k, v)

        logger.debug(f"returned property types: {propertyTypes}")
        return propertyTypes
        
    def _copyGraphQLString(graphQLString:str, copyGraphQLString:bool=False) -> None:
        """Can be applied to any core function to get the GraphQL string which is stored in the clipboard"""
        if copyGraphQLString == True and os.name == 'nt':
           return pyperclip.copy(graphQLString)

    def _getVariantId(variants:pd.DataFrame, name:str) -> str:
        """ Gets the variant Id from a given name""" 
        variants.set_index('name', inplace=True)
        return variants.loc[name][0]

    def _listGraphQlErrors(result:dict, key:str) -> None:
        """Print errors from GraphQL Query to log"""
        
        for i in result[key]['errors']:
            logger.error(i['message'])

    def _encodeBase64(file:str):
        with open(file) as file:
            content = file.read()
            content = base64.b64encode(content.encode('ascii'))
            return content.decode('UTF8')

    # deprecated:
    def _getInventoryId(self, inventoryName):
        inventory = self.inventories(where=f'name eq "{inventoryName}"')
        
        if inventory.empty:
            msg = f"Unknown inventory '{inventoryName}'."
            if self.raiseException: raise Exception(msg)
            else:
                logger.error(msg)
                return

        inventoryId = inventory.loc[0, 'inventoryId']
        logger.debug(f'returned inventory id: {inventoryId}')
        return inventoryId        

    def _arrayItemsToString(arrayItems:list, operation:str, cascadeDelete:bool) -> str:
        """Converts a list of array items to a graphQL string"""

        cDelValue = 'true' if cascadeDelete == True else 'false'

        if operation == 'insert':
            _arrayItems = 'insert: [\n'
            for item in arrayItems:
                _arrayItems += f'{{value: "{item}"}}\n'
            _arrayItems += ']'
            return _arrayItems
        if operation == 'removeByIndex':
            _arrayItems = f'cascadeDelete: {cDelValue}\n'
            _arrayItems += 'removeByIndex: ['
            for item in arrayItems:
                _arrayItems += f'{item}, '
            _arrayItems += ']'
            return _arrayItems
        if operation == 'removeById':
            _arrayItems = f'cascadeDelete: {cDelValue}\n'
            _arrayItems += 'removeById: ['
            for item in arrayItems:
                _arrayItems += f'"{item}", '
            _arrayItems += ']'
            return _arrayItems
        if operation == 'removeAll':
            _arrayItems = f'cascadeDelete: {cDelValue}\n'
            _arrayItems += 'removeByIndex: ['
            for item in arrayItems:
                _arrayItems += f'{item}, '
            _arrayItems += ']'
            return _arrayItems

    def _executeGraphQL(self, graphQLString):
        """Executes GraphQl, this code is used in every main function"""

        Utils._copyGraphQLString(graphQLString, self.defaults.copyGraphQLString)      
        logger.trace(f"GraphQLString: {graphQLString}")
        try:
            query = gql(graphQLString)
        except Exception as err:
            if self.raiseException: raise Exception(err)
            else:
                logger.error(err)
                return

        try:
            result = self.client.execute(query)
        except Exception as err:
            if self.raiseException: raise Exception(err)
            else:
                logger.error(err)
                return
        
        return result

    def _argNone(arg, value, enum=False) -> str:
        """
        Creates a simple string (to be embedded in graphQlString) 
        for arguments that are None by default.
        """
        if value == None:
            return ''
        else:
            if enum == True:
                return f'{arg}: {value}'
            else:
                if type(value) == str:
                    return f'{arg}: "{value}"'
                elif type(value) == float:
                    return f'{arg}: {value}'
                elif type(value) == int:
                    return f'{arg}: {value}'
                elif type(value) == bool:
                    if value == True:
                        return f'{arg}: true'
                    else:
                        return f'{arg}: false'
                else:
                    return f'{arg}: "{str(value)}"'
